
package com.locad.viewport;

import com.locad.geometry.Matrix4f;
import com.locad.geometry.Point3f;
import com.locad.geometry.Vec4f;
import java.awt.Point;
import org.apache.log4j.Logger;

public final class Camera
{
    private final Logger log;

    private static final float Epsilon = 1.0e-5f;

    public Point3f prevPos = new Point3f(0, 0, 0);
    public Point3f pos = new Point3f(-500, -500, -500);
    public Point3f lookAt = new Point3f(0f, 0f, 0f);

    private float yaw = 0f;
    private float roll = 0f;
    private float pitch = 0f;

    private float pan_sensitivity = 1.0f;

    private final Object matrixLock = new Object();
    public Matrix4f LastRot = Matrix4f.identityMatrix();
    public Matrix4f ThisRot = Matrix4f.identityMatrix();

    public Point startMousePt = new Point();
    public Point3f startVec = new Point3f(0f, 0f, 0f);          //Saved click vector
    public Point3f endVec = new Point3f(0f, 0f, 0f);          //Saved drag vector
    public float adjustWidth = 0f;       //Mouse bounds width
    public float adjustHeight = 0f;      //Mouse bounds height

    public Camera (Point3f pos, Point3f lookAt, float roll, float yaw, float pitch, float width, float height)
    {
        log = Logger.getLogger(getClass().getName());

        this.pos = pos;
        this.lookAt = lookAt;
        this.roll = roll;
        this.yaw = yaw;
        this.pitch = pitch;

        setBounds(width, height);
    }

    public Point3f getPos() { return pos; }
    public float getRoll() { return roll; }
    public float getYaw() { return yaw; }
    public float getPitch() { return pitch; }
    public float getSensitivity_Pan() { return pan_sensitivity; }

    public void setBounds(float newWidth, float newHeight)
    {
        assert ((newWidth > 1.0f) && (newHeight > 1.0f));

        //Set adjustment factor for width/height
        adjustWidth = 1.0f / ((newWidth - 1.0f) * 0.5f);
        adjustHeight = 1.0f / ((newHeight - 1.0f) * 0.5f);
    }

    public Point3f mapToSphere(Point point)
    {
        //Adjust point coords and scale down to range of [-1 ... 1]
        float x = (point.x * this.adjustWidth) - 1.0f;
        float y = 1.0f - (point.y * this.adjustHeight);

        //Compute the square of the length of the vector to the point from the center
        float length = (x * x) + (y * y);

        //If the point is mapped outside of the sphere... (length > radius squared)
        if (length > 1.0f)
        {
            //Compute a normalizing factor (radius / sqrt(length))
            float norm = (float) (1.0 / Math.sqrt(length));

            //Return the "normalized" vector, a point on the sphere
            return new Point3f(x * norm, -y * norm, 0f);
        }
        else    //Else it's on the inside
        {
            //Return a vector to a point mapped inside the sphere sqrt(radius squared - length)
            return new Point3f(x, -y, (float) Math.sqrt(1.0f - length));
        }
    }

    public void reset()
    {
        synchronized (matrixLock)
        {
            prevPos = pos.copy();
        }
    }

    public void startDrag(Point newPt)
    {
        // set start vector from mouse down event
        synchronized (matrixLock)
        {
            prevPos = pos.copy();
        }
        startVec = mapToSphere(newPt);
        startMousePt.x = newPt.x;
        startMousePt.y = newPt.y;
    }

    public void rotate(Point newPt)
    {
        //Map the point to the sphere
        endVec = mapToSphere(newPt);

        //Compute the vector perpendicular to the begin and end vectors
        Point3f perp = startVec.crossProd(endVec);

        // make sure the vector is greater than zero length
        if (perp.mag() > Epsilon)
        {
            //w is cosine (theta / 2), where theta is rotation angle
            Vec4f newRot = new Vec4f(perp.x, perp.y, perp.z, -startVec.dotProd(endVec));

            synchronized (matrixLock)
            {
                pos = new Matrix4f(newRot).times(prevPos);
            }
        }
    }

    public void pan(Point newPt)
    {
        lookAt.x -= pan_sensitivity * (startMousePt.x - newPt.getX());
        lookAt.z -= pan_sensitivity * (startMousePt.y - newPt.getY());
        startMousePt.x = newPt.x;
        startMousePt.y = newPt.y;
    }

    public void zoom(int scrollAmount, int wheelRotation)
    {
        float x = pos.x * 0.1f * scrollAmount * wheelRotation;
        float y = pos.y * 0.1f * scrollAmount * wheelRotation;
        float z = pos.z * 0.1f * scrollAmount * wheelRotation;

        pos.x += x;
        pos.y += y;
        pos.z += z;
    }
}
