
package com.locad.viewport;

import com.locad.enumerations.Constants;
import com.locad.geometry.Matrix4f;
import com.locad.geometry.Point3f;
import com.locad.geometry.Vec4f;
import com.locad.ldraw.ColorLibrary;
import com.locad.ldraw.Part;
import com.locad.ldraw.PartsLibrary;
import com.sun.opengl.util.BufferUtil;
import java.awt.Color;
import java.nio.IntBuffer;
import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.glu.GLU;
import org.apache.log4j.Logger;

public class Renderer implements GLEventListener
{
    private final Logger log;

    private GLU glu = new GLU();

    private Scene scene;
    private Camera camera;

    private float mouseX = 0;
    private float mouseY = 0;
    private boolean mouseClicked = false;
    
    private double fov = 3.0;
    private double nearClip = 100.0;
    private double farClip = 100000.0;

    private final Color bgColor;

    public Renderer (Part part, PartsLibrary partLib, ColorLibrary colorLib, Color bgColor)
    {
        log = Logger.getLogger(getClass().getName());

        this.bgColor = bgColor;

        scene = new Scene(part, partLib, colorLib);

        Point3f point = scene.getFurthestPointFromOrigin(part, part.orientation);
        float distX = 2 * point.z / (float) Math.tan(fov * Constants.Deg2Rad);
        float distY = 2 * point.x / (float) Math.tan(fov * Constants.Deg2Rad);
        float distZ = 2 * point.y / (float) Math.tan(fov * Constants.Deg2Rad);
        
        Point3f cameraPos = new Point3f(-distX, -distY, -distZ);
        Point3f cameraDir = new Point3f(0f, 0f, 0f);
        camera = new Camera(cameraPos, cameraDir, 0f, 0f, 0f, 500, 500);
    }

    public void init(GLAutoDrawable glDrawable)
    {
        final GL gl = glDrawable.getGL();

        camera.ThisRot = Matrix4f.identityMatrix();
        camera.LastRot = Matrix4f.identityMatrix();

        gl.glShadeModel (GL.GL_FLAT);

        gl.glClearColor(bgColor.getRed() / 255f, bgColor.getGreen() / 255f, bgColor.getBlue() / 255, bgColor.getAlpha() / 255f);
        gl.glClearDepth(1.0f);
        
        gl.glEnable(GL.GL_DEPTH_TEST);
        gl.glDepthFunc(GL.GL_LEQUAL);
        gl.glHint(GL.GL_PERSPECTIVE_CORRECTION_HINT, GL.GL_NICEST);

        gl.glPolygonMode(GL.GL_FRONT, GL.GL_FILL);

//        gl.glCullFace(GL.GL_BACK);
//        gl.glEnable(GL.GL_CULL_FACE);

        Light light0 = new Light(new float[] {0.06f, 0.06f, 0.06f, 1.0f}, new float[] {0.01f, 0.01f, 0.01f, 1.0f}, new float[] {0.01f, 0.01f, 0.01f, 1.0f}, new Vec4f(5000f, 0f, 0f, 1f));
        light0.ApplyToGL(gl, GL.GL_LIGHT0);

        Light light1 = new Light(new float[] {0.06f, 0.06f, 0.06f, 1.0f}, new float[] {0.001f, 0.001f, 0.001f, 1.0f}, new float[] {0.01f, 0.01f, 0.01f, 1.0f}, new Vec4f(0f, 5000f, 0f, 1f));
        light1.ApplyToGL(gl, GL.GL_LIGHT1);

        Light light2 = new Light(new float[] {0.06f, 0.06f, 0.06f, 1.0f}, new float[] {0.001f, 0.001f, 0.001f, 1.0f}, new float[] {0.01f, 0.01f, 0.01f, 1.0f}, new Vec4f(0f, 0f, 5000f, 1f));
        light2.ApplyToGL(gl, GL.GL_LIGHT2);

        Light light3 = new Light(new float[] {0.06f, 0.06f, 0.06f, 1.0f}, new float[] {0.001f, 0.001f, 0.001f, 1.0f}, new float[] {0.01f, 0.01f, 0.01f, 1.0f}, new Vec4f(-5000f, 0f, 0f, 1f));
        light3.ApplyToGL(gl, GL.GL_LIGHT3);

        Light light4 = new Light(new float[] {0.06f, 0.06f, 0.06f, 1.0f}, new float[] {0.001f, 0.001f, 0.001f, 1.0f}, new float[] {0.01f, 0.01f, 0.01f, 1.0f}, new Vec4f(0f, -5000f, 0f, 1f));
        light4.ApplyToGL(gl, GL.GL_LIGHT4);

        Light light5 = new Light(new float[] {0.06f, 0.06f, 0.06f, 1.0f}, new float[] {0.001f, 0.001f, 0.001f, 1.0f}, new float[] {0.01f, 0.01f, 0.01f, 1.0f}, new Vec4f(0f, 0f, -5000f, 1f));
        light5.ApplyToGL(gl, GL.GL_LIGHT5);

        scene.createScene(gl);

        Listener listener = new Listener(scene, camera);
        glDrawable.addKeyListener(listener);
        glDrawable.addMouseListener(listener);
        glDrawable.addMouseMotionListener(listener);
        glDrawable.addMouseWheelListener(listener);
    }

    public void display(GLAutoDrawable glDrawable)
    {
        final GL gl = glDrawable.getGL();

        gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glLoadIdentity();

        // Camera
        //glu.gluLookAt(camera.getX() + Math.cos(camera.getYaw()) * Math.cos(camera.getPitch()), camera.getY() + Math.sin(camera.getYaw()) * Math.cos(camera.getPitch()), camera.getZ() + Math.sin(camera.getPitch()), camera.getX(), camera.getY(), camera.getZ(), 0.0, -1.0, 0.0);
        glu.gluLookAt(camera.pos.x, camera.pos.y, camera.pos.z,
                      camera.lookAt.x, camera.lookAt.y, camera.lookAt.z,
                      0.0, -1.0, 0.0);

        // Render scene by calling display list
        gl.glCallList(scene.scene_dl);

        if (mouseClicked)
        {
            doSelection(gl);
            mouseClicked = false;
        }

        // Flush OpenGL pipeline
        gl.glFlush();
    }

    public void doSelection(GL gl)
    {
        int BUFSIZE = 512;
        IntBuffer selectBuffer = BufferUtil.newIntBuffer(BUFSIZE);

        // The Size Of The Viewport. [0] Is <x>, [1] Is <y>, [2] Is <length>, [3] Is <width>
        int[] viewport = new int[4];

        // This Sets The Array <viewport> To The Size And Location Of The Screen Relative To The Window
        gl.glGetIntegerv(GL.GL_VIEWPORT, viewport, 0);
        gl.glSelectBuffer(BUFSIZE, selectBuffer); // Tell OpenGL To Use Our Array For Selection

        // Puts OpenGL In Selection Mode. Nothing Will Be Drawn.  Object ID's and Extents Are Stored In The Buffer.
        gl.glRenderMode(GL.GL_SELECT);
        gl.glInitNames(); // Initializes The Name Stack
        gl.glPushName(0); // Push 0 (At Least One Entry) Onto The Stack

        // Create A Matrix That Will Zoom Up To A Small Portion Of The Screen, Where The Mouse Is.
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glPushMatrix();
        gl.glLoadIdentity();
        glu.gluPickMatrix((double) mouseX, (double) (viewport[3] - mouseY), 1.0f, 1.0f, viewport, 0);

        // Apply The Perspective Matrix
        glu.gluPerspective(45.0f, (float) (viewport[2] - viewport[0]) / (float) (viewport[3] - viewport[1]), 0.1f, 100.0f);
        gl.glMatrixMode(GL.GL_MODELVIEW);
        //drawTargets(gl); // Render The Targets To The Selection Buffer
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glPopMatrix();

        // switch back to modelview matrix
        gl.glMatrixMode(GL.GL_MODELVIEW);

        // get number of objects that were visible in this tiny view port (and switch back to render mode)
        int hits = gl.glRenderMode(GL.GL_RENDER);
    }

    public void displayChanged(GLAutoDrawable glDrawable, boolean modeChanged, boolean deviceChanged)
    {
        //init(glDrawable);
    }

    public void reshape(GLAutoDrawable glDrawable, int x, int y, int w, int h)
    {
        final GL gl = glDrawable.getGL();
        gl.glViewport(0, 0, w, h);
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();
        glu.gluPerspective(fov, (double) w / (double) h, nearClip, farClip);
        camera.setBounds((float) w, (float) h);
    }
}
