
package com.locad.ldraw;

import com.locad.enumerations.Enumerations.BFCTypes;
import com.locad.enumerations.Enumerations.MaterialTypes;
import com.locad.enumerations.Enumerations.ShapeTypes;
import com.locad.enumerations.Enumerations.ViewAngleModes;
import com.locad.geometry.*;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import org.apache.log4j.Logger;

public class Part extends Shape
{
    private final Logger log;

    private Matrix4f orientation = Matrix4f.identityMatrix();

    private String fileName = "";
    private String description = "";
    private String name = "";
    private String author = "";

    private boolean BFCflag = false;
    private BFCTypes bfc = BFCTypes.NONE;

    private List<Shape> shapes = new ArrayList<Shape>();

    public Part ()
    {
        shapeType = ShapeTypes.PART;
        log = Logger.getLogger(getClass().getName());
    }

    public Matrix4f getOrientation() { return orientation; }
    public String getFileName() { return fileName; }
    public String getDescripton() { return description; }
    public String getName() { return name; }
    public String getAuthor() { return author; }
    public boolean getBFCflag() { return BFCflag; }
    public BFCTypes getBFCType() { return bfc; }
    public List<Shape> getShapes() { return shapes; }

    @Override
    public void setOrientation(Matrix4f orientation)
    {
        this.orientation = orientation;
        for (Shape shape : shapes)
        {
            shape.setOrientation(orientation);
        }
    }

    public static Part from_File (String fileName)
    {
        boolean BFCFlag = false;

        Part part = new Part();
        part.fileName = fileName.replace("\\", "/").toUpperCase();

        try
        {
            Scanner s = new Scanner(new BufferedReader(new FileReader(fileName)));

            // read description (always first line)
            if (s.hasNextLine())
            {
                part.description = s.nextLine();
            }

            while (s.hasNextLine())
            {
                String line = s.nextLine();

                if (line.contains("0 Name"))
                {
                    part.name = line;
                }
                else if (line.contains("0 Author"))
                {
                    part.author = line;
                }
                else if(line.equals("0 BFC CERTIFY CW"))
                {
                    part.bfc = BFCTypes.CW;
                }
                else if (line.equals("0 BFC CERTIFY CCW"))
                {
                    part.bfc = BFCTypes.CCW;
                }

                if(line.equals("0 BFC INVERTNEXT"))
                {
                    BFCFlag = true;
                }
                String[] tokens = line.trim().split(" ");
                String[] t = new String[tokens.length];
                int ndx = 0;
                for (int j=0; j<tokens.length; j++)
                {
                    if (tokens[j] == null ? "" != null : !tokens[j].equals(""))
                    {
                        t[ndx] = tokens[j];
                        ndx++;
                    }
                }
                if (t.length > 1)
                {
                    if(t[0].equals("1"))
                    {
                        part.shapes.add(Part.from_PartEmbeded(PartEmbeded.from_Line(line, part.orientation, BFCFlag)));
                        BFCFlag = false;
                    }
                    else if(t[0].equals("2"))
                    {
                        part.shapes.add(Line.from_Line(line, part.orientation, part.bfc == BFCTypes.CW));
                    }
                    else if (t[0].equals("3"))
                    {
                        part.shapes.add(Triangle.from_Line(line, part.orientation, part.bfc == BFCTypes.CW));
                    }
                    else if (t[0].equals("4"))
                    {
                        part.shapes.add(Quadrilateral.from_Line(line, part.orientation, part.bfc == BFCTypes.CW));
                    }
                    else if (t[0].equals("5"))
                    {
                        //part.shapes.add(LineConditional.from_Line(line, part.orientation, part.bfc == BFCTypes.CW));
                    }
                }
            }
            s.close();
        }
        catch (FileNotFoundException ex)
        {
            part.log.error("error: part file name: " + part.fileName + " reason " + ex);
        }

        // trim the leading 0s and such
        if (part.description.length() > 2 && part.description.charAt(0) == '0')
        {
            part.description = part.description.substring(2);
        }
        if (part.name.length() > 7 && part.name.charAt(0) == '0')
        {
            part.name = part.name.substring(7).trim();
        }
        if (part.author.length() > 8 && part.author.charAt(0) == '0')
        {
            part.author = part.author.substring(8).trim();
        }

        //part.log.info(part);
        return part;
    }

    private static Part from_PartEmbeded(PartEmbeded p)
    {
        Part part = PartsLibrary.getInstance().getPart(p.fileName);
        part.colorCode = p.colorCode;
        part.setOrientation(p.orientation);
        part.BFCflag = p.BFCflag;
        return part;
    }

    @Override
    public void setColor(int requestedColorCode, int currentColorCode)
    {
        for (Shape shape : this.shapes)
        {
            shape.setColor(requestedColorCode, currentColorCode);
        }
    }

    @Override
    public void addToList(Map<MaterialTypes, ArrayList<Shape>> materials, int requestedColorCode, float lineWidth)
    {
        for (Shape shape : this.shapes)
        {
            if (shape instanceof Part)
            {
                Part p = (Part) shape;
                p.setColor(requestedColorCode, p.colorCode);

                if (p.colorCode == 16)
                    p.addToList(materials, requestedColorCode, lineWidth);
                else
                    p.addToList(materials, p.colorCode, lineWidth);
            }
            else
            {
                shape.addToList(materials, requestedColorCode, lineWidth);
            }
        }
    }
    
    @Override
    public void addToListBFC(Map<MaterialTypes, ArrayList<Shape>> materials, int requestedColorCode, float lineWidth, boolean BFCflag)
    {
        for (Shape shape : shapes)
        {
            boolean localBFC = BFCflag;

            if (shape instanceof Part)
            {
                Part p = (Part) shape;
                p.setColor(requestedColorCode, p.colorCode);

                if (p.BFCflag)
                    localBFC = !BFCflag;

                if (p.colorCode == 16)
                    p.addToListBFC(materials, requestedColorCode, lineWidth, localBFC);
                else
                    p.addToListBFC(materials, p.colorCode, lineWidth, localBFC);
            }
            else
            {
                shape.addToListBFC(materials, requestedColorCode, lineWidth, BFCflag);
            }
        }
    }

    @Override
    public Point3f getCentroid (Point3f min, Point3f max)
    {
        minPoint = new Point3f(Float.MAX_VALUE, Float.MAX_VALUE, Float.MAX_VALUE);
        maxPoint = new Point3f(-Float.MAX_VALUE, -Float.MAX_VALUE, -Float.MAX_VALUE);

        for (Shape shape : this.shapes)
        {
            centroid = shape.getCentroid(minPoint, maxPoint);
        }
        
        if (minPoint.x < min.x)
            min.x = minPoint.x;
        if (maxPoint.x > max.x)
            max.x = maxPoint.x;
        if (minPoint.y < min.y)
            min.y = minPoint.y;
        if (maxPoint.y > max.y)
            max.y = maxPoint.y;
        if (minPoint.z < min.z)
            min.z = minPoint.z;
        if (maxPoint.z > max.z)
            max.z = maxPoint.z;

        centroid = new Point3f(maxPoint.x - (maxPoint.x - minPoint.x) / 2f, maxPoint.y - (maxPoint.y - minPoint.y) / 2f, maxPoint.z - (maxPoint.z - minPoint.z) / 2f);

        return centroid;
    }

    @Override
    public boolean contains (float x, float y, ViewAngleModes viewAngle)
    {
        for (Shape shape : this.shapes)
        {
            if (shape.contains(x, y, viewAngle))
            {
                return true;
            }
        }
        return false;
    }

    @Override
    public Part copy()
    {
        Part that = new Part();
        that.fileName = this.fileName;
        that.description = this.description;
        that.name = this.name;
        that.author = this.author;
        that.bfc = this.bfc;
        that.BFCflag = this.BFCflag;
        that.orientation = this.orientation.copy();

        that.color = this.color;
        that.colorCode = this.colorCode;
        that.material = this.material;

        for (Shape shape : this.shapes)
        {
            that.shapes.add(shape.copy());
        }

        return that;
    }

    @Override
    public String toString ()
    {
        StringBuilder strBuffer = new StringBuilder();

        strBuffer.append("fileName ").append(fileName).append("\n");
        strBuffer.append("description ").append(description).append("\n");
        strBuffer.append("name ").append(name).append("\n");
        strBuffer.append("author ").append(author).append("\n");
        strBuffer.append("bfc ").append(bfc).append("\n");
        strBuffer.append("orientation ").append(orientation.toString()).append("\n");
        strBuffer.append(super.toString());

        for (Shape shape : shapes)
        {
            strBuffer.append(shape.toString()).append("\n");
        }

        return strBuffer.toString();
    }
}

