
package com.locad.ldraw;

import java.io.*;
import java.util.*;
import org.apache.log4j.Logger;

public class PartsLibrary
{
    private final Logger log;

    public List<String[]> primitiveDirs;
    public List<String[]> partDirs;

    private Map<String, Part> parts = new HashMap<String, Part>();
    public final Map<String, String> partPaths = new HashMap<String, String>();
    public final Map<String, String> partDescriptions = new HashMap<String, String>();

    private static PartsLibrary instance = null;

    public static PartsLibrary getInstance()
    {
        synchronized (PartsLibrary.class)
        {
            if (instance == null)
            {
                instance = new PartsLibrary();
            }
        }
        return instance;
    }

    public PartsLibrary ()
    {
        log = Logger.getLogger(getClass().getName());
    }

    public void readPartFiles (List<String[]> primitiveDirs, List<String[]> partDirs, String ldrawDir, String partListFile)
    {
        this.primitiveDirs = primitiveDirs;
        this.partDirs = partDirs;

        // primitives
        for (String[] dir : primitiveDirs)
        {
            scanDir(partPaths, dir[0], dir[1]);
        }
        
        // parts
        for (String[] dir : partDirs)
        {
            scanDir(partPaths, dir[0], dir[1]);
        }

        if (fileExists(partListFile))
        {
            readPartList(partListFile);
        }
        else
        {
            createPartList(partListFile);
        }
    }

    private void scanDir (Map<String, String> partPaths, String parentDir, String folderName)
    {
        File dir = new File(parentDir + folderName);
        String[] chld = dir.list();
        if (chld != null)
        {
            for (int i = 0; i < chld.length; i ++)
            {
                String filePath = parentDir + folderName + chld[i];
                String fileName = folderName + chld[i];
                filePath = filePath.replace("\\", "/").toUpperCase();
                fileName = fileName.replace("\\", "/").toUpperCase();
                if (partPaths.get(fileName) != null)
                {
                    log.warn("Duplicate part file found: " + filePath + " and " + partPaths.get(fileName).toString());
                }
                else
                {
                    partPaths.put(fileName, filePath);
                }
            }
        }
    }

    private void readPartList(String partListFile)
    {
        try
        {
            Scanner s = new Scanner(new BufferedReader(new FileReader(partListFile)));

            while (s.hasNextLine())
            {
                String[] tokens = s.nextLine().trim().split(" ");
                String[] t = new String[tokens.length];
                int ndx = 0;
                for (int j=0; j<tokens.length; j++)
                {
                    if (tokens[j] == null ? "" != null : !tokens[j].equals(""))
                    {
                        t[ndx] = tokens[j];
                        ndx++;
                    }
                }
                if (t.length > 1)
                {
                    partDescriptions.put(t[0], t[1]);
                }
            }
            s.close();
        }
        catch (FileNotFoundException ex)
        {
            log.error("error: part list file name: " + partListFile + " reason " + ex);
        }
    }

    private void createPartList(String partListFile)
    {
        Map<String, String> output = new HashMap<String, String>();
        for (String[] partDir : primitiveDirs)
        {
            List<String> partNames = getPartNamesFromPath(partDir[0] + partDir[1]);
            for (int i=0; i<partNames.size(); i++)
            {
                try
                {
                    Part part = getPart(partNames.get(i));
                    String description = part.getDescripton();
                    if (description.charAt(0) == '~')
                    {
                        description = description.substring(1);
                    }
                    output.put(part.getName(), description);
                }
                catch (Exception e)
                {
                    log.error(e);
                }
            }
        }

        // sort alphabetically based on description
        List<String> keys = new ArrayList(output.keySet());
        List<String> values = new ArrayList(output.values());
        TreeSet sortedSet = new TreeSet(values);
        Object[] sortedArray = sortedSet.toArray();

        // get the longest fileName
        int longest = 0;
        for (int i = 0; i < sortedArray.length; i++)
        {
            if (keys.get(values.indexOf(sortedArray[i])).length() > longest)
            {
                longest = keys.get(values.indexOf(sortedArray[i])).length();
            }
        }

        try
        {
            BufferedWriter out = new BufferedWriter(new FileWriter(partListFile));

            // write the data to file
            for (int i = 0; i < sortedArray.length; i++)
            {
                partDescriptions.put(keys.get(values.indexOf(sortedArray[i])), sortedArray[i].toString());
                out.write(keys.get(values.indexOf(sortedArray[i])));
                for (int j=0; j<(longest+5)-keys.get(values.indexOf(sortedArray[i])).length(); j++)
                {
                    out.write(" ");
                }
                out.write(sortedArray[i] + "\n");
            }

            out.close();
        }
        catch (Exception ex)
        {
            log.error("Could not create Primitives.lst file: " + ex);
        }
    }

    public String[] getFileNames()
    {
        return partPaths.keySet().toArray(new String[0]);
    }

    public List<String> getPartNamesFromPath(String path)
    {
        path = path.replace("\\", "/").toUpperCase();

        List<String> partNames = new ArrayList<String>();
        for (String fileName : partPaths.keySet())
        {
            if (partPaths.get(fileName).contains(path))
            {
                if (! partPaths.get(fileName).substring(path.length()).contains("/") && fileName.contains(".DAT"))
                {
                    partNames.add(fileName);
                }
            }
        }
        return partNames;
    }

    public Part getPart(String primitiveName)
    {
        if (primitiveName.isEmpty() || primitiveName == null)
            return null;

        primitiveName = primitiveName.replace("\\", "/").toUpperCase();

        if (!primitiveName.contains(".DAT") && !primitiveName.contains(".LDR"))
            return null;

        if (parts.get(primitiveName) != null)
        {
            return parts.get(primitiveName).copy();
        }
        else
        {
            if (partPaths.containsKey(primitiveName))
            {
                log.debug("creating new part: " + primitiveName);
                Part part = Part.from_File(partPaths.get(primitiveName));
                parts.put(primitiveName, part.copy());
                return part;
            }
            else
            {
                for (String fileName : partPaths.keySet())
                {
                    if (partPaths.get(fileName).contains(primitiveName))
                    {
                        log.debug("creating new part: " + primitiveName);
                        Part part = Part.from_File(partPaths.get(fileName));
                        parts.put(primitiveName, part.copy());
                        return part;
                    }
                }
                log.warn("not found in list primitiveName: " + primitiveName);
                return null;
            }
        }
    }

    private boolean fileExists (String filename)
    {
        //filename = filename.replace("/", File.pathSeparator);
        File file = new File(filename);
        return file.exists();
    }

    @Override
    public String toString ()
    {
        StringBuilder strBuffer = new StringBuilder();

        for (String name : partPaths.keySet())
        {
            strBuffer.append(name).append(" ").append(partPaths.get(name).toString()).append("\n");
        }

        for (String name : partPaths.keySet())
        {
            strBuffer.append(name).append(" ").append(partPaths.get(name).toString()).append("\n");
        }

        return strBuffer.toString();
    }
}
