
package com.locad.viewport;

import com.locad.enumerations.Enumerations.ViewAngleModes;
import com.locad.geometry.Matrix4f;
import com.locad.geometry.Point3f;
import com.locad.geometry.Vec4f;
import java.awt.Point;
import org.apache.log4j.Logger;

public class Camera
{
    private final Logger log;

    public ViewAngleModes viewAngle;

    private static final float Epsilon = 1.0e-5f;

    public Point3f prevPos = new Point3f(0, 0, 0);
    public Point3f pos = new Point3f(-500, -500, -500);
    public Point3f lookAt = new Point3f(0, 0, 0);

    public float yaw = 0f;
    public float roll = 0f;
    public float pitch = 0f;

    private final Object matrixLock = new Object();
    public Matrix4f LastRot = Matrix4f.identityMatrix();
    public Matrix4f ThisRot = Matrix4f.identityMatrix();

    public Point startMousePt = new Point();
    public Point3f startVec = new Point3f(0f, 0f, 0f);          //Saved click vector
    public Point3f endVec = new Point3f(0f, 0f, 0f);          //Saved drag vector
    public float width = 0f;       //Mouse bounds width
    public float height = 0f;      //Mouse bounds height

    public float radius = 0;
    public float left = 0;
    public float right = 0;
    public float bottom = 0;
    public float top = 0;
    public float near = 0;
    public float far = 0;

    public Camera (ViewAngleModes viewAngle)
    {
        log = Logger.getLogger(getClass().getName());

        this.viewAngle = viewAngle;
    }

    public void calcViewBox(Point3f centroid, Point3f minPoint, Point3f maxPoint)
    {
        Point3f thickness = new Point3f((maxPoint.x - minPoint.x) / 2f, (maxPoint.y - minPoint.y) / 2f, (maxPoint.z - minPoint.z) / 2f);

        radius = thickness.mag();
        lookAt = centroid;

        if(viewAngle == ViewAngleModes.FRONT)
        {
            left = -Math.max(thickness.x, thickness.y);
            right = Math.max(thickness.x, thickness.y);
            bottom = -Math.max(thickness.x, thickness.y);
            top = Math.max(thickness.x, thickness.y);
            near = 1f;
            far = 2*radius+2;
            roll = -1f;
            pitch = 0f;
            yaw = 0f;
            pos.x = lookAt.x;
            pos.y = lookAt.y;
            pos.z = lookAt.z-radius-2;
        }
        else if (viewAngle == ViewAngleModes.BACK)
        {
            left = -Math.max(thickness.x, thickness.y);
            right = Math.max(thickness.x, thickness.y);
            bottom = -Math.max(thickness.x, thickness.y);
            top = Math.max(thickness.x, thickness.y);
            near = 1f;
            far = 2*radius+2;
            roll = -1f;
            pitch = 0f;
            yaw = 0f;
            pos.x = lookAt.x;
            pos.y = lookAt.y;
            pos.z = lookAt.z+radius+2;
        }
        else if(viewAngle == ViewAngleModes.LEFT)
        {
            left = -Math.max(thickness.y, thickness.z);
            right = Math.max(thickness.y, thickness.z);
            bottom = -Math.max(thickness.y, thickness.z);
            top = Math.max(thickness.y, thickness.z);
            near = 1f;
            far = 2*radius+2;
            roll = -1f;
            pitch = 0f;
            yaw = 0f;
            pos.x = lookAt.x-radius-2;
            pos.y = lookAt.y;
            pos.z = lookAt.z;
        }
        else if(viewAngle == ViewAngleModes.RIGHT)
        {
            left = -Math.max(thickness.y, thickness.z);
            right = Math.max(thickness.y, thickness.z);
            bottom = -Math.max(thickness.y, thickness.z);
            top = Math.max(thickness.y, thickness.z);
            near = 1f;
            far = 2*radius+2;
            roll = -1f;
            pitch = 0f;
            yaw = 0f;
            pos.x = lookAt.x+radius+2;
            pos.y = lookAt.y;
            pos.z = lookAt.z;
        }
        else if(viewAngle == ViewAngleModes.TOP)
        {
            left = -Math.max(thickness.x, thickness.z);
            right = Math.max(thickness.x, thickness.z);
            bottom = -Math.max(thickness.x, thickness.z);
            top = Math.max(thickness.x, thickness.z);
            near = 1f;
            far = 2*radius+2;
            roll = 0f;
            pitch = 1f;
            yaw = 0f;
            pos.x = lookAt.x;
            pos.y = lookAt.y-radius-2;
            pos.z = lookAt.z;
        }
        else if(viewAngle == ViewAngleModes.BOTTOM)
        {
            left = -Math.max(thickness.x, thickness.z);
            right = Math.max(thickness.x, thickness.z);
            bottom = -Math.max(thickness.x, thickness.z);
            top = Math.max(thickness.x, thickness.z);
            near = 1f;
            far = 2*radius+2;
            roll = 0f;
            pitch = 1f;
            yaw = 0f;
            pos.x = lookAt.x;
            pos.y = lookAt.y+radius+2;
            pos.z = lookAt.z;
        }
        else
        {
            left = -radius;
            right = radius;
            bottom = -radius;
            top = radius;
            near = 1f;
            far = 3*radius+2;
            roll = -1f;
            pitch = 0f;
            yaw = 0f;
            pos.x = lookAt.x+radius+2;
            pos.y = lookAt.y-radius-2;
            pos.z = lookAt.z-radius-2;
        }
        //log.info("Pos " + pos + " centroid " + centroid + " left " + left + " bottom " + bottom + " far " + far);
    }

    public void setBounds(float width, float height)
    {
        this.width = width;
        this.height = height;
    }

    public void startDrag(Point newPt)
    {
        synchronized (matrixLock)
        {
            prevPos = pos.copy();
        }
        startMousePt.x = newPt.x;
        startMousePt.y = newPt.y;
        startVec = mapToSphere(newPt);
    }

    public Point3f mapToSphere(Point point)
    {
        //Set adjustment factor for width/height
        float adjustWidth = 1.0f / ((width - 1.0f) * 0.5f);
        float adjustHeight = 1.0f / ((height - 1.0f) * 0.5f);

        //Adjust point coords and scale down to range of [-1 ... 1]
        float x = (point.x * adjustWidth) - 1.0f;
        float y = 1.0f - (point.y * adjustHeight);

        //Compute the square of the length of the vector to the point from the center
        float length = (x * x) + (y * y);

        //If the point is mapped outside of the sphere... (length > radius squared)
        if (length > 1.0f)
        {
            //Compute a normalizing factor (radius / sqrt(length))
            float norm = (float) (1.0 / Math.sqrt(length));

            //Return the "normalized" vector, a point on the sphere
            return new Point3f(x * norm, -y * norm, norm);
        }
        else    //Else it's on the inside
        {
            //Return a vector to a point mapped inside the sphere sqrt(radius squared - length)
            return new Point3f(x, -y, (float) Math.sqrt(1.0f - length));
        }
    }

    public void rotate(Point newPt)
    {
        // TODO: bad things happen when the camera is straight up or straight down
        if (viewAngle == ViewAngleModes.ISOMETRIC || viewAngle == ViewAngleModes.PARTS_BAG)
        {
            //Map the point to the sphere
            endVec = mapToSphere(newPt);

            //Compute the vector perpendicular to the begin and end vectors
            Point3f perp = startVec.crossProd(endVec);

            // make sure the vector is greater than zero length
            if (perp.mag() > Epsilon)
            {
                //w is cosine (theta / 2), where theta is rotation angle
                Vec4f newRot = new Vec4f(perp.x, perp.y, perp.z, -startVec.dotProd(endVec));

                synchronized (matrixLock)
                {
                    pos = new Matrix4f(newRot).times(prevPos);
                    far = 3*pos.distanceTo(lookAt);
                }
            }
        }
    }

    public void pan(Point newPt)
    {
        if (viewAngle == ViewAngleModes.FRONT)
        {
            lookAt.x += (startMousePt.x - newPt.getX()) / width * (right - left);
            lookAt.y += (startMousePt.y - newPt.getY()) / height * (top - bottom);
            pos.x += (startMousePt.x - newPt.getX()) / width * (right - left);
            pos.y += (startMousePt.y - newPt.getY()) / height * (top - bottom);
        }
        else if(viewAngle == ViewAngleModes.BACK)
        {
            lookAt.x -= (startMousePt.x - newPt.getX()) / width * (right - left);
            lookAt.y += (startMousePt.y - newPt.getY()) / height * (top - bottom);
            pos.x -= (startMousePt.x - newPt.getX()) / width * (right - left);
            pos.y += (startMousePt.y - newPt.getY()) / height * (top - bottom);
        }
        else if (viewAngle == ViewAngleModes.RIGHT)
        {
            lookAt.z += (startMousePt.x - newPt.getX()) / width * (right - left);
            lookAt.y += (startMousePt.y - newPt.getY()) / height * (top - bottom);
            pos.z += (startMousePt.x - newPt.getX()) / width * (right - left);
            pos.y += (startMousePt.y - newPt.getY()) / height * (top - bottom);
        }
        else if (viewAngle == ViewAngleModes.LEFT)
        {
            lookAt.z -= (startMousePt.x - newPt.getX()) / width * (right - left);
            lookAt.y += (startMousePt.y - newPt.getY()) / height * (top - bottom);
            pos.z -= (startMousePt.x - newPt.getX()) / width * (right - left);
            pos.y += (startMousePt.y - newPt.getY()) / height * (top - bottom);
        }
        else if (viewAngle == ViewAngleModes.TOP)
        {
            lookAt.x += (startMousePt.x - newPt.getX()) / width * (right - left);
            lookAt.z -= (startMousePt.y - newPt.getY()) / height * (top - bottom);
            pos.x += (startMousePt.x - newPt.getX()) / width * (right - left);
            pos.z -= (startMousePt.y - newPt.getY()) / height * (top - bottom);
        }
        else if (viewAngle == ViewAngleModes.BOTTOM)
        {
            lookAt.x -= (startMousePt.x - newPt.getX()) / width * (right - left);
            lookAt.z -= (startMousePt.y - newPt.getY()) / height * (top - bottom);
            pos.x -= (startMousePt.x - newPt.getX()) / width * (right - left);
            pos.z -= (startMousePt.y - newPt.getY()) / height * (top - bottom);
        }
        else
        {
            // TODO: create a plane normal to the camera pos that will move the lookAt pos
            lookAt.x -= (startMousePt.x - newPt.getX()) / width * (right - left);
            lookAt.y -= (startMousePt.y - newPt.getY()) / height * (top - bottom);
            lookAt.z -= (startMousePt.y - newPt.getY()) / height * (top - bottom);
            pos.x -= (startMousePt.x - newPt.getX()) / width * (right - left);
            pos.y -= (startMousePt.y - newPt.getY()) / height * (top - bottom);
            pos.z -= (startMousePt.y - newPt.getY()) / height * (top - bottom);
        }
        startMousePt.x = newPt.x;
        startMousePt.y = newPt.y;
    }

    public void zoom(int scrollAmount, int wheelRotation)
    {
        float aspect = width / height;

        float oldWidth = (right - left);
        float newWidth = oldWidth * 0.4f * scrollAmount;

        left += (oldWidth - newWidth) * wheelRotation / 2f;
        right -= (oldWidth - newWidth) * wheelRotation / 2f;

        float oldHeight = (top - bottom);
        float newHeight = newWidth / aspect;

        bottom += (oldHeight - newHeight) * wheelRotation / 2f;
        top -= (oldHeight - newHeight) * wheelRotation / 2f;
    }
}
