
package com.locad.viewport;

import com.locad.enumerations.Enumerations.MaterialTypes;
import com.locad.geometry.Point3f;
import com.locad.geometry.Shape;
import com.locad.ldraw.Part;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import javax.media.opengl.GL;
import org.apache.log4j.Logger;

public class Scene
{
    private final Logger log;

    private boolean enableBFC;
    private int scene_dl;
    private Part part;
    private final float lineWidth;
    private Point3f minPoint = new Point3f(Float.MAX_VALUE, Float.MAX_VALUE, Float.MAX_VALUE);
    private Point3f maxPoint = new Point3f(-Float.MAX_VALUE, -Float.MAX_VALUE, -Float.MAX_VALUE);
    private Point3f centroid = new Point3f(0, 0, 0);
    private Map<MaterialTypes, ArrayList<Shape>> materials = new HashMap<MaterialTypes, ArrayList<Shape>>();

    public Scene(Part part, float lineWidth, boolean enableBFC)
    {
        log = Logger.getLogger(getClass().getName());

        this.part = part;
        this.lineWidth = lineWidth;
        this.enableBFC = enableBFC;

        for (MaterialTypes materialType : MaterialTypes.values())
        {
            materials.put(materialType, new ArrayList<Shape>());
        }

        if (part != null)
        {
            if (enableBFC)
                part.addToListBFC(materials, part.colorCode, lineWidth, false);
            else
                part.addToList(materials, part.colorCode, lineWidth);

            setCentroid();
        }
    }

    public boolean getEnableBFC() { return enableBFC; }
    public int getScene_dl() { return scene_dl; }
    public Part getPart() { return part; }
    public float getLineWidth() { return lineWidth; }
    public Point3f getCentroid() { return centroid; }
    public Point3f getMinPoint() { return minPoint; }
    public Point3f getMaxPoint() { return maxPoint; }

    public void createScene(GL gl)
    {
        setCentroid();

        scene_dl = gl.glGenLists(1);
        gl.glNewList(scene_dl, GL.GL_COMPILE);

        gl.glLineWidth(lineWidth);

        if (enableBFC)
        {
            for (MaterialTypes materialType : MaterialTypes.values())
            {
                setMaterialBFC(materialType, gl);
                for (Shape shape : materials.get(materialType))
                {
                    shape.addToGLwithBFC(gl);
                }
            }
        }
        else
        {
            for (MaterialTypes materialType : MaterialTypes.values())
            {
                setMaterial(materialType, gl);
                for (Shape shape : materials.get(materialType))
                {
                    shape.addToGL(gl);
                }
            }
        }
        
        gl.glEndList();
    }

    private void setMaterialBFC(MaterialTypes materialType, GL gl)
    {
        if (materialType == MaterialTypes.LUMINANT)
        {
            gl.glEnable(GL.GL_LIGHTING);
            gl.glDisable(GL.GL_BLEND);
            gl.glEnable(GL.GL_COLOR_MATERIAL);
            //gl.glMaterialfv(GL.GL_FRONT, GL.GL_EMISSION, blankMaterial, 0);
            Material material = new Material(new float[] {0.0f, 0.0f, 0.0f, 1.0f}, new float[] {0.5f, 0.5f, 0.5f, 1.0f}, new float[] {0.7f, 0.7f, 0.7f, 1.0f}, 0.1f);
            material.ApplyToGL(gl);
        }
        else if (materialType == MaterialTypes.RUBBER)
        {
            gl.glEnable(GL.GL_LIGHTING);
            gl.glDisable(GL.GL_BLEND);
            gl.glEnable(GL.GL_COLOR_MATERIAL);
            Material material = new Material(new float[] {0.2f, 0.2f, 0.2f, 1.0f}, new float[] {-0.2f, -0.2f, -0.2f, 1.0f}, new float[] {-0.2f, -0.2f, -0.2f, 1.0f}, 0.0f);
            material.ApplyToGL(gl);
        }
        else if (materialType == MaterialTypes.PLASTIC)
        {
            gl.glEnable(GL.GL_LIGHTING);
            gl.glDisable(GL.GL_BLEND);
            gl.glEnable(GL.GL_COLOR_MATERIAL);
            Material material = new Material(new float[] {0.0f, 0.0f, 0.0f, 1.0f}, new float[] {0.5f, 0.5f, 0.5f, 1.0f}, new float[] {0.5f, 0.5f, 0.5f, 1.0f}, 0.3f);
            material.ApplyToGL(gl);
        }
        else if (materialType == MaterialTypes.GLITTER)
        {
            gl.glEnable(GL.GL_LIGHTING);
            gl.glDisable(GL.GL_BLEND);
            gl.glEnable(GL.GL_COLOR_MATERIAL);
            Material material = new Material(new float[] {0.4f, 0.4f, 0.4f, 1.0f}, new float[] {0.9f, 0.9f, 0.9f, 1.0f}, new float[] {0.9f, 0.9f, 0.9f, 1.0f}, 0.4f);
            material.ApplyToGL(gl);
        }
        else if (materialType == MaterialTypes.SPECKLE)
        {
            gl.glEnable(GL.GL_LIGHTING);
            gl.glDisable(GL.GL_BLEND);
            gl.glEnable(GL.GL_COLOR_MATERIAL);
            Material material = new Material(new float[] {0.0f, 0.0f, 0.0f, 1.0f}, new float[] {0.5f, 0.5f, 0.5f, 1.0f}, new float[] {0.7f, 0.7f, 0.7f, 1.0f}, 0.1f);
            material.ApplyToGL(gl);
        }
        else if (materialType == MaterialTypes.METAL)
        {
            gl.glEnable(GL.GL_LIGHTING);
            gl.glDisable(GL.GL_BLEND);
            gl.glEnable(GL.GL_COLOR_MATERIAL);
            Material material = new Material(new float[] {0.3f, 0.3f, 0.3f, 1.0f}, new float[] {0.8f, 0.8f, 0.8f, 1.0f}, new float[] {0.8f, 0.8f, 0.8f, 1.0f}, 0.2f);
            material.ApplyToGL(gl);
        }
        else if (materialType == MaterialTypes.PEARLESCENT)
        {
            gl.glEnable(GL.GL_LIGHTING);
            gl.glDisable(GL.GL_BLEND);
            gl.glEnable(GL.GL_COLOR_MATERIAL);
            Material material = new Material(new float[] {0.1f, 0.1f, 0.1f, 1.0f}, new float[] {0.7f, 0.7f, 0.7f, 1.0f}, new float[] {0.7f, 0.7f, 0.7f, 1.0f}, 0.2f);
            material.ApplyToGL(gl);
        }
        else if (materialType == MaterialTypes.CHROME)
        {
            gl.glEnable(GL.GL_LIGHTING);
            gl.glDisable(GL.GL_BLEND);
            gl.glEnable(GL.GL_COLOR_MATERIAL);
            Material material = new Material(new float[] {0.4f, 0.4f, 0.4f, 1.0f}, new float[] {0.9f, 0.9f, 0.9f, 1.0f}, new float[] {0.9f, 0.9f, 0.9f, 1.0f}, 0.4f);
            material.ApplyToGL(gl);
        }
        else if (materialType == MaterialTypes.TRANSPARENT)
        {
            gl.glDisable(GL.GL_LIGHTING);
            gl.glEnable(GL.GL_BLEND);
            gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
            Material material = new Material(new float[] {0.0f, 0.0f, 0.0f, 1.0f}, new float[] {0.5f, 0.5f, 0.5f, 1.0f}, new float[] {0.7f, 0.7f, 0.7f, 1.0f}, 0.1f);
            material.ApplyToGL(gl);
        }
    }

    private void setMaterial (MaterialTypes materialType, GL gl)
    {
        if (materialType == MaterialTypes.TRANSPARENT)
        {
            gl.glEnable(GL.GL_BLEND);
            gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
            gl.glEnable(GL.GL_COLOR_MATERIAL);
            Material material = new Material(new float[] {0.0f, 0.0f, 0.0f, 1.0f}, new float[] {0.5f, 0.5f, 0.5f, 1.0f}, new float[] {0.7f, 0.7f, 0.7f, 1.0f}, 0.3f);
            material.ApplyToGL(gl);
        }
        else
        {
            gl.glDisable(GL.GL_BLEND);
            gl.glEnable(GL.GL_COLOR_MATERIAL);
            Material material = new Material(new float[] {0.0f, 0.0f, 0.0f, 1.0f}, new float[] {0.5f, 0.5f, 0.5f, 1.0f}, new float[] {0.5f, 0.5f, 0.5f, 1.0f}, 0.3f);
            material.ApplyToGL(gl);
        }
    }

    private void setCentroid ()
    {
        minPoint = new Point3f(Float.MAX_VALUE, Float.MAX_VALUE, Float.MAX_VALUE);
        maxPoint = new Point3f(-Float.MAX_VALUE, -Float.MAX_VALUE, -Float.MAX_VALUE);

        if (part != null)
        {
            for (Shape shape : part.getShapes())
            {
                centroid = shape.getCentroid(minPoint, maxPoint);
            }

            centroid.x = maxPoint.x - (maxPoint.x - minPoint.x) / 2f;
            centroid.y = maxPoint.y - (maxPoint.y - minPoint.y) / 2f;
            centroid.z = maxPoint.z - (maxPoint.z - minPoint.z) / 2f;
        }
        else
        {
            centroid = new Point3f(0,0,0);
            minPoint = new Point3f(0,0,0);
            maxPoint = new Point3f(0,0,0);
        }
    }
}
