
package com.locad.components;

import com.locad.enumerations.Enumerations.PartCategories;
import com.locad.enumerations.Enumerations.ViewAngleModes;
import com.locad.events.PartsTreeEvent;
import com.locad.geometry.Matrix4f;
import com.locad.geometry.Point3f;
import com.locad.ldraw.Part;
import com.locad.ldraw.PartDescription;
import com.locad.ldraw.PartsLibrary;
import com.locad.ldraw.Project;
import com.locad.main.Configuration;
import com.locad.model.Model;
import com.locad.viewport.Canvas;
import java.awt.Dimension;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.*;
import javax.swing.*;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import org.apache.log4j.Logger;

public class PartsBagComponent
{
    private final Logger log;

    private final Configuration config = Configuration.getInstance();
    private final PartsLibrary partsLib = PartsLibrary.getInstance();

    public Canvas partsBag;
    public JSplitPane partsPanel;

    private PartsTreeEvent currentState = new PartsTreeEvent("PartsTreeEvent");

    private int numPartsBagRows = 0;
    private int numPartsBagCols = 0;

    private ExecutorService executorService;

    public PartsBagComponent()
    {
        log = Logger.getLogger(getClass().getName());

        init();
    }

    private void init()
    {
        JTabbedPane partsPanelTab = new JTabbedPane();
        partsPanelTab.setTabLayoutPolicy(1);

        partsPanelTab.add(createTree(PartCategories.PARTS), 0);
        partsPanelTab.setTitleAt(0, PartCategories.PARTS.toString());

        partsPanelTab.add(createTree(PartCategories.SUB_PARTS), 1);
        partsPanelTab.setTitleAt(1, PartCategories.SUB_PARTS.toString());

        partsPanelTab.add(createTree(PartCategories.PRIMITIVES), 2);
        partsPanelTab.setTitleAt(2, PartCategories.PRIMITIVES.toString());

        partsPanelTab.add(createTree(PartCategories.HI_RES_PRIMITIVES), 3);
        partsPanelTab.setTitleAt(3, PartCategories.HI_RES_PRIMITIVES.toString());

        currentState.selectedCategory = PartCategories.PARTS;
        currentState.selectedFolder = "A";

        partsBag = new Canvas(ViewAngleModes.PARTS_BAG);
        partsBag.setPreferredSize(new Dimension(numPartsBagCols * 10, numPartsBagRows * 52 + 92));

        JScrollPane partsBagScroll = new JScrollPane(partsBag);
        partsBagScroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
        partsBagScroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);

        partsPanel = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
        partsPanel.setTopComponent(partsPanelTab);
        partsPanel.setBottomComponent(partsBagScroll);
        partsPanel.setDividerLocation(300);
    }
    
    public void postInit()
    {
        partsBag.postInit();
        
        Project project = createPartsBagProject();
        partsBag.setProject(project);
        partsBag.update();
        partsBag.setPreferredSize(new Dimension(numPartsBagCols * 10, numPartsBagRows * 52 + 92));
        partsPanel.setDividerLocation(partsPanel.getDividerLocation() + 1);
        partsPanel.setDividerLocation(partsPanel.getDividerLocation() - 1);
        partsBag.repaint();
        partsPanel.repaint();        
    }

    private JScrollPane createTree(PartCategories category)
    {
        Map<String, ArrayList<String>> partTreeNames = new HashMap<String, ArrayList<String>>();
        for (PartDescription d : partsLib.descriptions)
        {
            if (d.category == category)
            {
                if ( !partTreeNames.containsKey(d.description.substring(0, 1)))
                {
                    partTreeNames.put(d.description.substring(0, 1), new ArrayList<String>());
                }
                partTreeNames.get(d.description.substring(0, 1)).add(d.description);
            }
        }

        String[] letters = partTreeNames.keySet().toArray(new String[0]);
        Arrays.sort(letters);

        DefaultMutableTreeNode root = new DefaultMutableTreeNode(category.toString());
        for (int i=0; i<letters.length; i++)
        {
            DefaultMutableTreeNode node = new DefaultMutableTreeNode(letters[i]);
            root.add(node);

            String[] names = partTreeNames.get(letters[i]).toArray(new String[0]);
            //Arrays.sort(names);
            for (int j=0; j<names.length; j++)
            {
                DefaultMutableTreeNode child = new DefaultMutableTreeNode(names[j]);
                node.add(child);
            }
        }
        JTree tree = new JTree(root);
        tree.addTreeSelectionListener(new treeSelectionListener(tree));

//        DefaultTreeCellRenderer cellRenderer = new DefaultTreeCellRenderer();
//        cellRenderer.setOpenIcon(null);
//        cellRenderer.setClosedIcon(null);
//        cellRenderer.setLeafIcon(null);
//        tree.setCellRenderer(cellRenderer);

        JScrollPane partsTreeScroll = new JScrollPane(tree);
        partsTreeScroll.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
        partsTreeScroll.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);

        return partsTreeScroll;
    }

    private Project createPartsBagProject()
    {
        Project project = new Project("PartsBag");
        
        ArrayList<String> fileNames = new ArrayList<String>();
        for (PartDescription d : partsLib.descriptions)
        {
            if (d.category == currentState.selectedCategory && d.description.startsWith(currentState.selectedFolder))
            {
                fileNames.add(d.fileName);
            }
        }

        // kick off threads to read parts
        int ndx = 0;
        executorService = Executors.newFixedThreadPool(8);
        while (ndx < fileNames.size())
        {
            ArrayList<Future<Part>> parts = new ArrayList<Future<Part>>();
            for (int i=0; i<8; i++)
            {
                if (ndx < fileNames.size())
                {
                    parts.add(executorService.submit(new GetPart(fileNames.get(ndx))));
                }
                ndx++;
            }

            int i = 0;
            numPartsBagRows = 0;
            numPartsBagCols = 3;
            int vertical = 0;
            int horizontal = 0;
            for (Future<Part> result : parts)
            {
                try
                {
                    Part part = result.get();
                    if (part == null)
                    {
                        continue;
                    }

                    Point3f minPoint = new Point3f(Float.MAX_VALUE, Float.MAX_VALUE, Float.MAX_VALUE);
                    Point3f maxPoint = new Point3f(-Float.MAX_VALUE, -Float.MAX_VALUE, -Float.MAX_VALUE);
                    part.getBoundingBox(Matrix4f.angleViewMatrix(), minPoint, maxPoint);
                    float width = maxPoint.y - minPoint.y;
                    float height = maxPoint.z - minPoint.z;
                    float size = Math.max(width, height);
                    Matrix4f orientation = Matrix4f.angleViewMatrix();
                    //log.info(part.description + " " + width + " " + height + " " + size);
                    while (size > 100)
                    {
                        orientation = orientation.times(Matrix4f.identityMatrix().times(0.5f));
                        minPoint = new Point3f(Float.MAX_VALUE, Float.MAX_VALUE, Float.MAX_VALUE);
                        maxPoint = new Point3f(-Float.MAX_VALUE, -Float.MAX_VALUE, -Float.MAX_VALUE);
                        part.getBoundingBox(orientation, minPoint, maxPoint);
                        width = maxPoint.y - minPoint.y;
                        height = maxPoint.z - minPoint.z;
                        size = Math.max(width, height);
                    }
                    while (size < 100)
                    {
                        orientation = orientation.times(Matrix4f.identityMatrix().times(1.1f));
                        minPoint = new Point3f(Float.MAX_VALUE, Float.MAX_VALUE, Float.MAX_VALUE);
                        maxPoint = new Point3f(-Float.MAX_VALUE, -Float.MAX_VALUE, -Float.MAX_VALUE);
                        part.getBoundingBox(orientation, minPoint, maxPoint);
                        width = maxPoint.y - minPoint.y;
                        height = maxPoint.z - minPoint.z;
                        size = Math.max(width, height);
                    }

    //                minPoint = new Point3f(Float.MAX_VALUE, Float.MAX_VALUE, Float.MAX_VALUE);
    //                maxPoint = new Point3f(-Float.MAX_VALUE, -Float.MAX_VALUE, -Float.MAX_VALUE);
    //                part.getBoundingBox(orientation, minPoint, maxPoint);
    //                Point3f centroid = maxPoint.minus(maxPoint.minus(minPoint).div(2f));
                    //log.info(part.description + " " + width + " " + height + " " + size);

                    if (i % numPartsBagCols == 0)
                    {
                        numPartsBagRows++;
                        vertical += 100;
                        horizontal = 0;
                    }
                    else
                    {
                        horizontal += 100;
                    }

                    orientation.getMatrix()[12] = horizontal;// - centroid.x * sizeRatio;
                    orientation.getMatrix()[13] = vertical;// - centroid.y * sizeRatio;

                    project.addPart(part.fileName, orientation, config.defaultPartsBagColorCode);

        //            partsBag.getScene().setProject(project);
        //            partsBag.update();
        //            partsBag.setPreferredSize(new Dimension(numPartsBagCols * 10, numPartsBagRows * 52 + 92));
        //            partsPanel.setDividerLocation(partsPanel.getDividerLocation() + 1);
        //            partsPanel.setDividerLocation(partsPanel.getDividerLocation() - 1);
        //            partsBag.repaint();
        //            partsPanel.repaint();
                    i++;
                }
                catch (InterruptedException ie)
                {
                    log.error("interrupted", ie);
                }
                catch (ExecutionException ee)
                {
                    log.error("execution problem", ee);
                }
            }
        }

        return project;
    }

    public class treeSelectionListener implements TreeSelectionListener
    {
        private final JTree tree;

        public treeSelectionListener(JTree tree)
        {
            this.tree = tree;
        }

        public void valueChanged(TreeSelectionEvent e)
        {
            String selection = tree.getLastSelectedPathComponent().toString();
            if (selection.length() == 1)
            {
                currentState.selectedFolder = selection;
                Project project = null;//createPartsBagProject();

                partsBag.setProject(project);
                partsBag.display();
                partsBag.setPreferredSize(new Dimension(numPartsBagCols * 10, numPartsBagRows * 52 + 92));
                partsPanel.setDividerLocation(partsPanel.getDividerLocation() + 1);
            }
            else
            {
                currentState.selectedFileName = PartsLibrary.getInstance().getFileNameFromDescription(selection);
            }
        }
    }
    
    private class GetPart implements Callable<Part>
    {
        private final String fileName;

        public GetPart (String fileName)
        {
            this.fileName = fileName;
        }

        public Part call () throws Exception
        {
            return PartsLibrary.getInstance().getPart(fileName);
        }
    }
}
