
package com.locad.ldraw;

import com.locad.enumerations.Enumerations.BFCTypes;
import com.locad.enumerations.Enumerations.MaterialTypes;
import com.locad.enumerations.Enumerations.ViewAngleModes;
import com.locad.geometry.*;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.Scanner;
import javax.media.opengl.GL;
import org.apache.log4j.Logger;

public class Part extends Shape
{
    private static final Logger log = Logger.getLogger(Part.class);
    private final PartsLibrary partsLib = PartsLibrary.getInstance();
    public final String fileName;
    public final String description;
    public final ArrayList<Shape> shapes;

    public Part(String fileName, String description, ArrayList<Shape> shapes)
    {
        this.fileName = fileName;
        this.description = description;
        this.shapes = shapes;
    }

    public static Part from_File(String filePath, String fileName, int defaultColorCode)
    {
        String description = "";
        ArrayList<Shape> shapes = new ArrayList<Shape>();
        boolean BFCFlag = false;
        BFCTypes bfc = BFCTypes.NONE;

        try
        {
            Scanner s = new Scanner(new BufferedReader(new FileReader(filePath)));

            // read description (always first line)
            if (s.hasNextLine())
            {
                description = s.nextLine();
            }

            while (s.hasNextLine())
            {
                String line = s.nextLine();

                if (line.equals("0 BFC CERTIFY CW"))
                {
                    bfc = BFCTypes.CW;
                }
                else if (line.equals("0 BFC CERTIFY CCW"))
                {
                    bfc = BFCTypes.CCW;
                }

                if (line.equals("0 BFC INVERTNEXT"))
                {
                    BFCFlag = true;
                }
                line = line.replace("\t", " ");
                String[] tokens = line.trim().split(" ");
                String[] t = new String[tokens.length];
                int ndx = 0;
                for (int j = 0; j < tokens.length; j++)
                {
                    if (tokens[j] == null ? "" != null : !tokens[j].equals(""))
                    {
                        t[ndx] = tokens[j];
                        ndx++;
                    }
                }
                if (t.length > 1)
                {
                    if (t[0].equals("1"))
                    {
                        try
                        {
                            shapes.add(PartEmbeded.from_Line(line, BFCFlag));
                        }
                        catch (Exception ex)
                        {
                            log.error("Bad file format for file " + filePath + " from this line " + line + " cause: " + ex);
                        }
                        BFCFlag = false;
                    }
                    else if (t[0].equals("2"))
                    {
                        try
                        {
                            shapes.add(Line.from_Line(line, bfc == BFCTypes.CW));
                        }
                        catch (Exception ex)
                        {
                            log.error("Bad file format for file " + filePath + " from this line " + line + " cause: " + ex);
                        }
                    }
                    else if (t[0].equals("3"))
                    {
                        try
                        {
                            shapes.add(Triangle.from_Line(line, bfc == BFCTypes.CW));
                        }
                        catch (Exception ex)
                        {
                            log.error("Bad file format for file " + filePath + " from this line " + line + " cause: " + ex);
                        }
                    }
                    else if (t[0].equals("4"))
                    {
                        try
                        {
                            shapes.add(Quadrilateral.from_Line(line, bfc == BFCTypes.CW));
                        }
                        catch (Exception ex)
                        {
                            log.error("Bad file format for file " + filePath + " from this line " + line + " cause: " + ex);
                        }
                    }
                    else if (t[0].equals("5"))
                    {
                        //part.shapes.add(LineConditional.from_Line(line, part.orientation, part.bfc == BFCTypes.CW));
                    }
                }
            }
            s.close();
        }
        catch (FileNotFoundException ex)
        {
            log.error("error: part: " + filePath + " reason " + ex);
        }

        // trim the leading 0s and such
        if (description.length() > 2 && description.charAt(0) == '0')
        {
            description = description.substring(2);
        }

        //part.log.info(part);
        return new Part(fileName, description, shapes);
    }

    @Override
    public void getBoundingBox(Matrix4f orientation, Point3f minPoint, Point3f maxPoint)
    {
        for (Shape shape : shapes)
        {
            if (shape instanceof PartEmbeded)
            {
                PartEmbeded p = (PartEmbeded) shape;
                partsLib.getPart(p.fileName).getBoundingBox(p.orientation.times(orientation), minPoint, maxPoint);
            }
            else
            {
                shape.getBoundingBox(orientation, minPoint, maxPoint);
            }
        }
    }

    @Override
    public boolean contains(Matrix4f orientation, Point3f line, ViewAngleModes viewAngle)
    {
        for (Shape shape : shapes)
        {
            if (shape instanceof PartEmbeded)
            {
                PartEmbeded p = (PartEmbeded) shape;
                if (partsLib.getPart(p.fileName).contains(p.orientation.times(orientation), line, viewAngle))
                {
                    return true;
                }
            }
            else
            {
                if (shape.contains(orientation, line, viewAngle))
                {
                    //log.info(shape.toString(orientation));
                    return true;
                }
            }
        }
        return false;
    }

    @Override
    public void addToGL(GL gl, short requestedColorCode, MaterialTypes materialType)
    {
        for (Shape shape : shapes)
        {
            if (shape instanceof PartEmbeded)
            {
                PartEmbeded p = (PartEmbeded) shape;

                gl.glPushMatrix();
                gl.glMultMatrixf(p.orientation.getMatrix(), 0);

                if (p.getColorCode() == 16)
                {
                    partsLib.getPart(p.fileName).addToGL(gl, requestedColorCode, materialType);
                }
                else
                {
                    partsLib.getPart(p.fileName).addToGL(gl, p.getColorCode(), materialType);
                }

                gl.glPopMatrix();
            }
            else
            {
                shape.addToGL(gl, requestedColorCode, materialType);
            }
        }
    }

    @Override
    public void addToGLwithBFC(GL gl, short requestedColorCode, MaterialTypes materialType, boolean BFCflag)
    {
        boolean localBFC;

        for (Shape shape : shapes)
        {
            localBFC = BFCflag;
            if (shape instanceof PartEmbeded)
            {
                PartEmbeded p = (PartEmbeded) shape;
                gl.glPushMatrix();
                gl.glMultMatrixf(p.orientation.getMatrix(), 0);

                if (p.BFCflag)
                {
                    localBFC = !localBFC;
                }

                if (p.getColorCode() == 16)
                {
                    partsLib.getPart(p.fileName).addToGLwithBFC(gl, requestedColorCode, materialType, localBFC);
                }
                else
                {
                    partsLib.getPart(p.fileName).addToGLwithBFC(gl, p.getColorCode(), materialType, localBFC);
                }

                gl.glPopMatrix();
            }
            else
            {
                shape.addToGLwithBFC(gl, requestedColorCode, materialType, localBFC);
            }
        }
    }

    @Override
    public String toString()
    {
        StringBuilder strBuffer = new StringBuilder();

        strBuffer.append("\n");
        strBuffer.append("shapeType ").append("Part").append("\n");
        strBuffer.append("fileName ").append(fileName).append("\n");
        strBuffer.append("description ").append(description).append("\n");
        strBuffer.append(super.toString()).append("\n");

        for (Shape shape : shapes)
        {
            strBuffer.append(shape.toString()).append("\n");
        }

        return strBuffer.toString();
    }
}
