
package com.locad.ldraw;

import com.locad.enumerations.Enumerations.PartCategories;
import com.locad.main.Configuration;
import java.io.*;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Scanner;
import java.util.concurrent.*;
import org.apache.log4j.Logger;

public final class PartsLibrary
{
    private final Logger log;

    private final Configuration config = Configuration.getInstance();

    public ArrayList<PartDescription> descriptions = new ArrayList<PartDescription>();

    private static PartsLibrary instance = null;

    private ExecutorService executorService;

    public static PartsLibrary getInstance()
    {
        synchronized (PartsLibrary.class)
        {
            if (instance == null)
            {
                instance = new PartsLibrary();
            }
        }
        return instance;
    }

    private PartsLibrary()
    {
        log = Logger.getLogger(getClass().getName());
    }

    public void readPartFiles()
    {
        if (fileExists(config.primitivesListFile))
        {
            readPrimitivesList(config.primitivesListFile);
        }
        else
        {
            scanDir(config.partFolders.get(PartCategories.SUB_PARTS)[0], config.partFolders.get(PartCategories.SUB_PARTS)[1], false);
            scanDir(config.partFolders.get(PartCategories.PRIMITIVES)[0], config.partFolders.get(PartCategories.PRIMITIVES)[1], false);
            scanDir(config.partFolders.get(PartCategories.HI_RES_PRIMITIVES)[0], config.partFolders.get(PartCategories.HI_RES_PRIMITIVES)[1], false);

            createPrimitivesList(config.primitivesListFile);
        }
        
        if (fileExists(config.partsListFile))
        {
            readPartsList(config.partsListFile);
        }
        else
        {
            scanDir( config.partFolders.get(PartCategories.PARTS)[0], config.partFolders.get(PartCategories.PARTS)[1], false);

            createPartsList(config.partsListFile);
        }

        scanDir(config.partFolders.get(PartCategories.MODELS)[0], config.partFolders.get(PartCategories.MODELS)[1], true);
    }

    private void scanDir(String parentDir, String folderName, boolean recursive)
    {
        File dir = new File(parentDir + folderName);
        String[] child = dir.list();
        if (child != null)
        {
            for (int i = 0; i < child.length; i ++)
            {
                if (new File(parentDir + folderName + child[i]).isDirectory() && recursive)
                {
                    scanDir(parentDir + folderName, child[i] + "/", recursive);
                }
                String filePath = (parentDir + folderName + child[i]).replace("\\", "/").toUpperCase();
                String fileName = (folderName + child[i]).replace("\\", "/").toUpperCase();

                PartDescription description = new PartDescription();
                description.fileName = fileName;
                description.filePath = filePath;
                descriptions.add(description);
            }
        }
    }

    private void readPartsList(String partsListFile)
    {
        try
        {
            Scanner s = new Scanner(new BufferedReader(new FileReader(partsListFile)));

            while (s.hasNextLine())
            {
                String line = s.nextLine().replace("\\", "/").toUpperCase();
                int cut = line.indexOf(".DAT");
                cut += 4;
                if (cut > 0 && cut < line.length() -1)
                {
                    String fileName = line.substring(0, cut);
                    String description = line.substring(cut).trim();

                    PartDescription d = new PartDescription();
                    d.category = PartCategories.PARTS;
                    d.fileName = fileName;
                    d.filePath = (config.partFolders.get(PartCategories.PARTS)[0] + fileName).replace("\\", "/").toUpperCase();
                    d.description = description;
                    descriptions.add(d);
                }
            }
            s.close();
        }
        catch (FileNotFoundException ex)
        {
            log.error("error: part list file name: " + partsListFile + " reason " + ex);
        }
    }

    private void readPrimitivesList(String primitivesListFile)
    {
        try
        {
            Scanner s = new Scanner(new BufferedReader(new FileReader(primitivesListFile)));

            while (s.hasNextLine())
            {
                String line = s.nextLine().replace("\\", "/").toUpperCase();
                int cut = line.indexOf(".DAT");
                cut += 4;
                if (cut > 0 && cut < line.length() -1)
                {
                    String fileName = line.substring(0, cut);
                    String description = line.substring(cut).trim();

                    String filePath = (config.partFolders.get(PartCategories.PRIMITIVES)[0] + fileName).replace("\\", "/").toUpperCase();
                    PartCategories category = PartCategories.PRIMITIVES;
                    if (fileName.startsWith("S/"))
                    {
                        category = PartCategories.SUB_PARTS;
                        filePath = (config.partFolders.get(PartCategories.SUB_PARTS)[0] + fileName).replace("\\", "/").toUpperCase();
                    }
                    else if (fileName.startsWith("48/"))
                    {
                        category = PartCategories.HI_RES_PRIMITIVES;
                        filePath = (config.partFolders.get(PartCategories.HI_RES_PRIMITIVES)[0] + fileName).replace("\\", "/").toUpperCase();
                    }

                    PartDescription d = new PartDescription();
                    d.category = category;
                    d.fileName = fileName;
                    d.filePath = filePath;
                    d.description = description;
                    descriptions.add(d);
                }
            }
            s.close();
        }
        catch (FileNotFoundException ex)
        {
            log.error("error: part list file name: " + primitivesListFile + " reason " + ex);
        }
    }

    private void createPartsList(String partsListFile)
    {
        ArrayList<String> fileNames = getFileNamesFromPath(config.partFolders.get(PartCategories.PARTS)[0] + config.partFolders.get(PartCategories.PARTS)[1]);

        int numRead = 0;
        ArrayList<PartDescription> localDescriptions = new ArrayList<PartDescription>();
        executorService = Executors.newFixedThreadPool(Math.min(20, fileNames.size()));
        while(numRead < fileNames.size())
        {
            int reading = Math.min(20, fileNames.size() - numRead);
            ArrayList<Future<Part>> results = new ArrayList<Future<Part>>();
            for (int i=numRead; i<numRead + reading; i++)
            {
                results.add(executorService.submit(new GetPart(fileNames.get(i))));
            }

            for (Future<Part> result : results)
            {
                try
                {
                    Part part = result.get();

                    if (part.description == null || part.description.startsWith("_") || part.description.startsWith("~"))
                    {
                        continue;
                    }

                    PartDescription d = new PartDescription();
                    d.category = PartCategories.PARTS;
                    d.fileName = part.fileName;
                    d.description = part.description;
                    d.part = part;
                    localDescriptions.add(d);
                }
                catch (InterruptedException ie)
                {
                    //log.error("interrupted", ie);
                }
                catch (ExecutionException ee)
                {
                    //log.error("execution problem", ee);
                }
            }
            numRead += reading;
        }

        // sort alphabetically based on description
        Collections.sort(localDescriptions, new PartDescriptionComparator());

        // get the longest fileName
        int longest = 0;
        for (PartDescription d : localDescriptions)
        {
            if (d.fileName.length() > longest)
            {
                longest = d.fileName.length();
            }
        }

        try
        {
            BufferedWriter out = new BufferedWriter(new FileWriter(partsListFile));

            // write the data to file
            for (PartDescription d : localDescriptions)
            {
                out.write(d.fileName);
                for (int j=0; j<(longest+5)-d.fileName.length(); j++)
                {
                    out.write(" ");
                }
                out.write(d.description + "\n");
            }

            out.close();
        }
        catch (Exception ex)
        {
            log.error("Could not create " + partsListFile + " file: " + ex);
        }
    }

    private class GetPart implements Callable<Part>
    {
        private final String fileName;

        public GetPart(String fileName)
        {
            this.fileName = fileName;
        }

        public Part call() throws Exception
        {
            return getPart(fileName);
        }
    }

    public class CreateList implements Callable<Integer>
    {
        private final String[] folder;
        private final ArrayList<PartDescription> mapping;
        private final PartCategories category;

        public CreateList(String[] folder, ArrayList<PartDescription> mapping, PartCategories category)
        {
            this.folder = folder;
            this.mapping = mapping;
            this.category = category;
        }

        public Integer call() throws Exception
        {
            ArrayList<String> partNames = getFileNamesFromPath(folder[0] + folder[1]);
            for (int i=0; i<partNames.size(); i++)
            {
                Part part = getPart(partNames.get(i));
                
                if (part == null)
                {
                    continue;
                }

//                if (part.description.startsWith("_") || part.description.startsWith("~"))
//                {
//                    continue;
//                }

                PartDescription d = new PartDescription();
                d.category = category;
                d.fileName = part.fileName;
                d.description = part.description;
                d.part = part;
                mapping.add(d);
            }

            int longest = 0;
            for (PartDescription d : mapping)
            {
                if (d.fileName.length() > longest)
                {
                    longest = d.fileName.length();
                }
            }
            return longest;
        }
    }

    private void createPrimitivesList(String primitivesListFile)
    {
        executorService = Executors.newFixedThreadPool(5);
        ArrayList<Future<Integer>> results = new ArrayList<Future<Integer>>();

        results.add(executorService.submit(new CreateList(config.partFolders.get(PartCategories.SUB_PARTS), descriptions, PartCategories.SUB_PARTS)));
        results.add(executorService.submit(new CreateList(config.partFolders.get(PartCategories.PRIMITIVES), descriptions, PartCategories.PRIMITIVES)));
        results.add(executorService.submit(new CreateList(config.partFolders.get(PartCategories.HI_RES_PRIMITIVES), descriptions, PartCategories.HI_RES_PRIMITIVES)));

        int longest = 0;
        for (Future<Integer> result : results)
        {
            try
            {
                if (result.get() > longest)
                {
                    longest = result.get();
                }
            }
            catch (InterruptedException ie)
            {
                log.error("interrupted", ie);
            }
            catch (ExecutionException ee)
            {
                log.error("execution problem", ee);
            }
        }

        try
        {
            // write the data to file
            BufferedWriter out = new BufferedWriter(new FileWriter(primitivesListFile));

            // sort alphabetically based on description
            Collections.sort(descriptions, new PartDescriptionComparator());
            for (PartDescription d : descriptions)
            {
                out.write(d.fileName);
                for (int j=0; j<(longest+5)-d.fileName.length(); j++)
                {
                    out.write(" ");
                }
                out.write(d.description + "\n");
            }

            out.close();
        }
        catch (Exception ex)
        {
            log.error("Could not create " + primitivesListFile + "  file: " + ex);
        }
    }

    public class StringArrayComparator implements Comparator
    {
        public int compare(Object r1, Object r2)
        {
            return ((String[]) r1)[1].compareTo(((String[]) r2)[1]);
        }
    }

    public class PartDescriptionComparator implements Comparator
    {
        public int compare(Object r1, Object r2)
        {
            PartDescription d1 = (PartDescription) r1;
            PartDescription d2 = (PartDescription) r2;
            
            if (d1.category == d2.category)
            {
                return d1.description.compareTo(d2.description);
            }
            else
            {
                return d1.category.ordinal() - d2.category.ordinal();
            }
        }
    }

    public ArrayList<String> getFileNamesFromPath(String path)
    {
        path = path.replace("\\", "/").toUpperCase();

        ArrayList<String> partNames = new ArrayList<String>();
        for (PartDescription d : descriptions)
        {
            if (d.filePath.contains(path))
            {
                if (! d.filePath.substring(path.length()).contains("/") && d.fileName.contains(".DAT"))
                {
                    partNames.add(d.fileName);
                }
            }
        }
        return partNames;
    }

    public String getFileNameFromDescription(String description)
    {
        description = description.replace("\\", "/").toUpperCase();

        for (PartDescription d : descriptions)
        {
            if (d.filePath.equals(description))
            {
                return d.fileName;
            }
        }
        return null;
    }

    public Part getPart(String fileName)
    {
        if (fileName.isEmpty() || fileName == null)
            return null;

        fileName = fileName.replace("\\", "/").toUpperCase();

        if (!fileName.contains(".DAT") && !fileName.contains(".LDR"))
            return null;

        for (PartDescription d : descriptions)
        {
            if (d.fileName.equals(fileName))
            {
                if (d.part != null)
                {
                    return d.part;
                }
                else if (d.filePath != null)
                {
                    log.debug("creating new part: " + fileName);
                    d.part = Part.from_File(d.filePath, fileName, config.defaultColorCode);
                    return d.part;
                }
            }
        }

        if (fileExists((config.partFolders.get(PartCategories.PARTS)[0] + fileName).replace("\\", "/")))
        {
            PartDescription d = new PartDescription();
            d.category = PartCategories.PARTS;
            d.fileName = fileName;
            d.filePath = (config.partFolders.get(PartCategories.PARTS)[0] + fileName).replace("\\", "/").toUpperCase();
            d.part = Part.from_File(d.filePath, fileName, config.defaultColorCode);
            descriptions.add(d);
            return d.part;
        }

        if (fileExists((config.partFolders.get(PartCategories.SUB_PARTS)[0] + fileName).replace("\\", "/")))
        {
            PartDescription d = new PartDescription();
            d.category = PartCategories.SUB_PARTS;
            d.fileName = fileName;
            d.filePath = (config.partFolders.get(PartCategories.SUB_PARTS)[0] + fileName).replace("\\", "/").toUpperCase();
            d.part = Part.from_File(d.filePath, fileName, config.defaultColorCode);
            descriptions.add(d);
            return d.part;
        }

        if (fileExists((config.partFolders.get(PartCategories.PRIMITIVES)[0] + fileName).replace("\\", "/")))
        {
            PartDescription d = new PartDescription();
            d.category = PartCategories.PRIMITIVES;
            d.fileName = fileName;
            d.filePath = (config.partFolders.get(PartCategories.PRIMITIVES)[0] + fileName).replace("\\", "/").toUpperCase();
            d.part = Part.from_File(d.filePath, fileName, config.defaultColorCode);
            descriptions.add(d);
            return d.part;
        }

        if (fileExists((config.partFolders.get(PartCategories.HI_RES_PRIMITIVES)[0] + fileName).replace("\\", "/")))
        {
            PartDescription d = new PartDescription();
            d.category = PartCategories.HI_RES_PRIMITIVES;
            d.fileName = fileName;
            d.filePath = (config.partFolders.get(PartCategories.HI_RES_PRIMITIVES)[0] + fileName).replace("\\", "/").toUpperCase();
            d.part = Part.from_File(d.filePath, fileName, config.defaultColorCode);
            descriptions.add(d);
            return d.part;
        }

        for (PartDescription d : descriptions)
        {
            if (d.fileName.contains(fileName))
            {
                if (d.part != null)
                {
                    return d.part;
                }
                else if (d.filePath != null)
                {
                    log.debug("creating new part: " + fileName);
                    d.part = Part.from_File(d.filePath, fileName, config.defaultColorCode);
                    return d.part;
                }
            }
        }

        log.warn("not found in list fileName: " + fileName);
        return null;
    }

    private boolean fileExists(String filename)
    {
        File file = new File(filename);
        return file.exists();
    }

    @Override
    public String toString()
    {
        StringBuilder strBuffer = new StringBuilder();

        for (PartDescription d : descriptions)
        {
            strBuffer.append(d.fileName).append(" ").append(d.description).append("\n");
        }

        return strBuffer.toString();
    }
}
