
package com.locad.ldraw;

import com.locad.enumerations.Enumerations.BFCTypes;
import com.locad.enumerations.Enumerations.MaterialTypes;
import com.locad.enumerations.Enumerations.ViewAngleModes;
import com.locad.geometry.*;
import com.locad.main.Configuration;
import com.locad.model.ModelObject;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;
import javax.media.opengl.GL;

public class Project extends ModelObject
{
    private final Configuration config = Configuration.getInstance();
    private final PartsLibrary partsLib = PartsLibrary.getInstance();
    private String name = "";
    private String author = "";
    private boolean BFCflag = false;
    private BFCTypes bfc = BFCTypes.NONE;
    private String description = "";
    private List<Shape> shapes = new ArrayList<Shape>();
    private Point3f minPoint;
    private Point3f maxPoint;
    private Point3f centroid = new Point3f(0, 0, 0);

    public Project(String fileName)
    {
        super(fileName);
    }

    public String getFileName()
    {
        return uuid;
    }

    public String getName()
    {
        return name;
    }

    public String getAuthor()
    {
        return author;
    }

    public boolean getBFCflag()
    {
        return BFCflag;
    }

    public BFCTypes getBFCType()
    {
        return bfc;
    }

    public Point3f getCentroid()
    {
        return centroid;
    }

    public Point3f getMinPoint()
    {
        return minPoint;
    }

    public Point3f getMaxPoint()
    {
        return maxPoint;
    }

    public List<Shape> getShapes()
    {
        return shapes;
    }

    public static Project from_File(File filePath, short defaultColorCode)
    {
        boolean BFCFlag = false;

        Project project = new Project(filePath.getName());
        try
        {
            Scanner s = new Scanner(new BufferedReader(new FileReader(filePath)));

            // read description (always first line)
            if (s.hasNextLine())
            {
                project.description = s.nextLine();
            }

            while (s.hasNextLine())
            {
                String line = s.nextLine();

                if (line.contains("0 Name"))
                {
                    project.name = line;
                }
                else if (line.contains("0 Author"))
                {
                    project.author = line;
                }
                else if (line.equals("0 BFC CERTIFY CW"))
                {
                    project.bfc = BFCTypes.CW;
                }
                else if (line.equals("0 BFC CERTIFY CCW"))
                {
                    project.bfc = BFCTypes.CCW;
                }

                if (line.equals("0 BFC INVERTNEXT"))
                {
                    BFCFlag = true;
                }
                String[] tokens = line.trim().split(" ");
                String[] t = new String[tokens.length];
                int ndx = 0;
                for (int j = 0; j < tokens.length; j++)
                {
                    if (tokens[j] == null ? "" != null : !tokens[j].equals(""))
                    {
                        t[ndx] = tokens[j];
                        ndx++;
                    }
                }
                if (t.length > 1)
                {
                    if (t[0].equals("1"))
                    {
                        try
                        {
                            project.shapes.add(PartEmbeded.from_Line(line, BFCFlag));
                        }
                        catch (Exception ex)
                        {
                            project.log.error("Bad file format for file " + filePath + " from this line " + line + " cause: " + ex);
                        }
                        BFCFlag = false;
                    }
                    else if (t[0].equals("2"))
                    {
                        try
                        {
                            project.shapes.add(Line.from_Line(line, project.bfc == BFCTypes.CW));
                        }
                        catch (Exception ex)
                        {
                            project.log.error("Bad file format for file " + filePath + " from this line " + line + " cause: " + ex);
                        }
                    }
                    else if (t[0].equals("3"))
                    {
                        try
                        {
                            project.shapes.add(Triangle.from_Line(line, project.bfc == BFCTypes.CW));
                        }
                        catch (Exception ex)
                        {
                            project.log.error("Bad file format for file " + filePath + " from this line " + line + " cause: " + ex);
                        }
                    }
                    else if (t[0].equals("4"))
                    {
                        try
                        {
                            project.shapes.add(Quadrilateral.from_Line(line, project.bfc == BFCTypes.CW));
                        }
                        catch (Exception ex)
                        {
                            project.log.error("Bad file format for file " + filePath + " from this line " + line + " cause: " + ex);
                        }
                    }
                    else if (t[0].equals("5"))
                    {
                        //project.shapes.add(LineConditional.from_Line(line, project.orientation, project.bfc == BFCTypes.CW));
                    }
                }
            }
            s.close();
        }
        catch (FileNotFoundException ex)
        {
            project.log.error("error: project: " + filePath + " reason " + ex);
        }

        // trim the leading 0s and such
        if (project.description.length() > 2 && project.description.charAt(0) == '0')
        {
            project.description = project.description.substring(2);
        }
        if (project.name.length() > 7 && project.name.charAt(0) == '0')
        {
            project.name = project.name.substring(7).trim();
        }
        if (project.author.length() > 8 && project.author.charAt(0) == '0')
        {
            project.author = project.author.substring(8).trim();
        }

        //project.log.info(project);
        return project;
    }

    public void addPart(String fileName, Matrix4f orientation, short colorCode)
    {
        this.shapes.add(new PartEmbeded(fileName, orientation, colorCode, BFCflag, false));
    }

    public void getBoundingBox()
    {
        minPoint = new Point3f(Float.MAX_VALUE, Float.MAX_VALUE, Float.MAX_VALUE);
        maxPoint = new Point3f(-Float.MAX_VALUE, -Float.MAX_VALUE, -Float.MAX_VALUE);

        for (Shape shape : shapes)
        {
            if (shape instanceof PartEmbeded)
            {
                PartEmbeded p = (PartEmbeded) shape;
                partsLib.getPart(p.fileName).getBoundingBox(p.orientation.times(Matrix4f.identityMatrix()), minPoint, maxPoint);
            }
            else
            {
                shape.getBoundingBox(Matrix4f.identityMatrix(), minPoint, maxPoint);
            }
        }

        centroid = maxPoint.minus(maxPoint.minus(minPoint).div(2f));
    }

    public Shape getShape(Point3f line, ViewAngleModes viewAngle)
    {
        Shape minShape = null;

        float value = -Float.MAX_VALUE;
        if (viewAngle == ViewAngleModes.TOP || viewAngle == ViewAngleModes.LEFT || viewAngle == ViewAngleModes.FRONT)
        {
            value = Float.MAX_VALUE;
        }

        for (Shape shape : shapes)
        {
            Shape testShape = shape;
            Matrix4f orientation = Matrix4f.identityMatrix();
            if (shape instanceof PartEmbeded)
            {
                PartEmbeded p = (PartEmbeded) shape;
                testShape = partsLib.getPart(p.fileName);
                orientation = p.orientation.times(Matrix4f.identityMatrix());
//                if (partsLib.getPart(p.fileName).contains(p.orientation, line, viewAngle))
//                {
//                    log.info(shape.colorCode);
//                }
            }
            if (testShape.contains(orientation, line, viewAngle))
            {
                //log.info(shape.colorCode);

                Point3f min = new Point3f(Float.MAX_VALUE, Float.MAX_VALUE, Float.MAX_VALUE);
                Point3f max = new Point3f(-Float.MAX_VALUE, -Float.MAX_VALUE, -Float.MAX_VALUE);
                testShape.getBoundingBox(orientation, min, max);

                if (viewAngle == ViewAngleModes.FRONT)
                {
                    if (min.z <= value)
                    {
                        value = min.z;
                        minShape = shape;
                    }
                }
                else if (viewAngle == ViewAngleModes.BACK)
                {
                    if (max.z > value)
                    {
                        value = max.z;
                        minShape = shape;
                    }
                }
                else if (viewAngle == ViewAngleModes.RIGHT)
                {
                    if (max.x > value)
                    {
                        value = max.x;
                        minShape = shape;
                    }
                }
                else if (viewAngle == ViewAngleModes.LEFT)
                {
                    if (min.x <= value)
                    {
                        value = min.x;
                        minShape = shape;
                    }
                }
                else if (viewAngle == ViewAngleModes.TOP)
                {
                    if (min.y <= value)
                    {
                        value = min.y;
                        minShape = shape;
                    }
                }
                else if (viewAngle == ViewAngleModes.BOTTOM)
                {
                    if (max.y > value)
                    {
                        value = max.y;
                        minShape = shape;
                    }
                }
            }
        }

        //log.info(minShape.colorCode);
        return minShape;
    }

    public void addToGL(GL gl, MaterialTypes materialType)
    {
        for (Shape shape : shapes)
        {
            if (shape instanceof PartEmbeded)
            {
                PartEmbeded p = (PartEmbeded) shape;

                gl.glPushMatrix();
                gl.glMultMatrixf(p.orientation.getMatrix(), 0);

                if (p.getColorCode() == 16)
                {
                    partsLib.getPart(p.fileName).addToGL(gl, config.defaultColorCode, materialType);
                }
                else
                {
                    partsLib.getPart(p.fileName).addToGL(gl, p.getColorCode(), materialType);
                }

                gl.glPopMatrix();
            }
            else
            {
                shape.addToGL(gl, config.defaultColorCode, materialType);
            }
        }
    }

    public void addToGLwithBFC(GL gl, MaterialTypes materialType)
    {
        boolean localBFC;

        for (Shape shape : shapes)
        {
            localBFC = false;
            if (shape instanceof PartEmbeded)
            {
                PartEmbeded p = (PartEmbeded) shape;

                if (p.BFCflag)
                {
                    localBFC = !localBFC;
                }

                gl.glPushMatrix();
                gl.glMultMatrixf(p.orientation.getMatrix(), 0);

                Part part = partsLib.getPart(p.fileName);

                if (p.getColorCode() == 16)
                {
                    part.addToGLwithBFC(gl, config.defaultColorCode, materialType, localBFC);
                }
                else
                {
                    part.addToGLwithBFC(gl, p.getColorCode(), materialType, localBFC);
                }

                gl.glPopMatrix();
            }
            else
            {
                shape.addToGLwithBFC(gl, config.defaultColorCode, materialType, localBFC);
            }
        }
    }

    public Project copy()
    {
        Project that = new Project(uuid);
        that.description = this.description;
        that.author = this.author;
        that.bfc = this.bfc;
        that.BFCflag = this.BFCflag;

        for (Shape shape : this.shapes)
        {
            that.shapes.add(shape);
        }

        return that;
    }

    @Override
    public String toString()
    {
        StringBuilder strBuffer = new StringBuilder();

        strBuffer.append("\n");
        strBuffer.append("fileName ").append(uuid).append("\n");
        strBuffer.append("description ").append(description).append("\n");

        for (Shape shape : shapes)
        {
            strBuffer.append(shape.toString()).append("\n");
        }

        return strBuffer.toString();
    }
}
