
package com.locad.main;

import com.locad.components.PartsBagComponent;
import com.locad.enumerations.Enumerations.MaterialTypes;
import com.locad.enumerations.Enumerations.ViewAngleModes;
import com.locad.geometry.Shape;
import com.locad.ldraw.ColorLibrary;
import com.locad.ldraw.Project;
import com.locad.model.Model;
import com.locad.model.ModelListener;
import com.locad.viewport.Canvas;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.File;
import java.util.Arrays;
import javax.swing.*;
import org.apache.log4j.Logger;

public final class Desktop extends JFrame
{
    private final Logger log;

    private final Configuration config = Configuration.getInstance();

    private Project project;

    private Canvas front;
    private Canvas right;
    private Canvas top;
    private Canvas isometric;

    private JTabbedPane colorPanel;
    private PartsBagComponent partsComponent;

    private JSplitPane leftToolSplit;
    private JSplitPane rightToolSplit;
    private JSplitPane leftSplit;
    private JSplitPane rightSplit;
    private JSplitPane mainSplit;
    private JMenuBar menuBar;

    private JFileChooser fileChooser;

    private JDialog aboutBox;

    private static Desktop instance = null;

    public static Desktop getInstance()
    {
        synchronized (Desktop.class)
        {
            if (instance == null)
            {
                instance = new Desktop();
            }
        }
        return instance;
    }

    private Desktop()
    {
        log = Logger.getLogger(getClass().getName());

        createColorPanel();
        partsComponent = new PartsBagComponent();
        createViewPanel();
        createMenuBar();

        init();

        Model.getInstance().addListener(Project.class, new ProjectListener());
    }

    private void init()
    {
        setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
        this.setPreferredSize(new Dimension(1600, 1024));
        this.setSize(new Dimension(1600, 1024));

        leftSplit = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
        leftSplit.setTopComponent(front);
        leftSplit.setBottomComponent(top);
        leftSplit.setDividerLocation(this.getHeight() / 2);

        rightSplit = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
        rightSplit.setTopComponent(right);
        rightSplit.setBottomComponent(isometric);
        rightSplit.setDividerLocation(this.getHeight() / 2);

        mainSplit = new JSplitPane();
        mainSplit.setLeftComponent(leftSplit);
        mainSplit.setRightComponent(rightSplit);
        mainSplit.setDividerLocation((this.getWidth() - 400) / 2);

        SplitPaneSynchronizer sps = new SplitPaneSynchronizer(leftSplit, rightSplit);

        leftToolSplit = new JSplitPane();
        leftToolSplit.setDividerLocation(200);
        leftToolSplit.setLeftComponent(partsComponent.partsPanel);
        leftToolSplit.setRightComponent(mainSplit);

        rightToolSplit = new JSplitPane();
        rightToolSplit.setDividerLocation(this.getWidth() - 200);
        rightToolSplit.setLeftComponent(leftToolSplit);
        rightToolSplit.setRightComponent(colorPanel);

        this.setLayout(new BorderLayout());
        this.add(rightToolSplit, BorderLayout.CENTER);
        this.add(menuBar, BorderLayout.NORTH);
    }

    public void postInit()
    {
        front.postInit();
        right.postInit();
        top.postInit();
        isometric.postInit();
        partsComponent.postInit();
    }

    private class ProjectListener extends ModelListener<Project>
    {
        @Override
        public void add (Project p)
        {
            project = p;
            setTitle(project.getFileName());

            front.setProject(project);
            right.setProject(project);
            top.setProject(project);
            isometric.setProject(project);
        }

        @Override
        public void update (Project p)
        {
            project = p;
            setTitle(project.getFileName());

            front.setProject(project);
            right.setProject(project);
            top.setProject(project);
            isometric.setProject(project);
        }

        @Override
        public void remove(Project project)
        {
            project = null;
            setTitle("");

            front.setProject(null);
            right.setProject(null);
            top.setProject(null);
            isometric.setProject(null);
        }
    }

    private void createViewPanel()
    {
        front = new Canvas(ViewAngleModes.FRONT);
        right = new Canvas(ViewAngleModes.RIGHT);
        top = new Canvas(ViewAngleModes.TOP);
        isometric = new Canvas(ViewAngleModes.ISOMETRIC);
    }

    private void createColorPanel()
    {
        ColorLibrary colorLib = ColorLibrary.getInstance();

        Short[] colorCodes = colorLib.colors.keySet().toArray(new Short[0]);
        Arrays.sort(colorCodes);

        int longestName = 0;
        for (short i=0; i<colorCodes.length; i++)
        {
            if (colorLib.colors.get(colorCodes[i]).name.length() > longestName)
            {
                longestName = colorLib.colors.get(colorCodes[i]).name.length();
            }
        }

        colorPanel = new JTabbedPane();
        for (MaterialTypes materialType : MaterialTypes.values())
        {
            createColorSet(colorLib, colorCodes, longestName, materialType);
        }
        colorPanel.setTabLayoutPolicy(1);
        colorPanel.setSelectedIndex(MaterialTypes.PLASTIC.ordinal());
    }

    private void createColorSet(ColorLibrary colorLib, Short[] colorCodes, int longestName, MaterialTypes materialType)
    {
        int numColors = 0;
        JPanel colorList = new JPanel();
        for (short i=0; i<colorCodes.length; i++)
        {
            JPanel value = new JPanel();
            value.setPreferredSize(new Dimension(25, 25));
            value.setBackground(new Color(colorLib.colors.get(colorCodes[i]).fillColor.red + 128, colorLib.colors.get(colorCodes[i]).fillColor.green + 128, colorLib.colors.get(colorCodes[i]).fillColor.blue + 128));

            JLabel label = new JLabel();
            label.setFont(new Font("Arial Narrow", Font.BOLD, 12));
            label.setForeground(Color.BLACK);
            label.setText(" " + colorLib.colors.get(colorCodes[i]).name);

            JPanel color = new JPanel();
            color.setLayout(new BorderLayout());
            color.setPreferredSize(new Dimension(longestName*5 + 28, 25));
            color.add(value, BorderLayout.WEST);
            color.add(label, BorderLayout.CENTER);
            color.addMouseListener(new mouseListener(colorCodes[i]));

            if (colorLib.colors.get(colorCodes[i]).materialType == materialType)
            {
                numColors++;
                colorList.add(color);
            }
        }
        colorList.setPreferredSize(new Dimension(longestName*5 + 28, 30 * numColors));
        JScrollPane scrollPane = new JScrollPane(colorList);
        scrollPane.setVerticalScrollBarPolicy(ScrollPaneConstants.VERTICAL_SCROLLBAR_ALWAYS);
        scrollPane.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);

        colorPanel.add(scrollPane, materialType.ordinal());
        colorPanel.setTitleAt(materialType.ordinal(), materialType.toString());
    }

    public class mouseListener implements MouseListener
    {
        private final short colorCode;

        public mouseListener(short colorCode)
        {
            this.colorCode = colorCode;
        }

        public void mouseClicked(MouseEvent e)
        {
            for (Shape shape : project.getShapes())
            {
                if (shape.getIsSelected())
                {
                    shape.setColorCode(colorCode);
                }
            }
            Model.getInstance().merge(project);
        }
        public void mousePressed(MouseEvent e)
        {
        }
        public void mouseReleased(MouseEvent e)
        {
        }
        public void mouseEntered(MouseEvent e)
        {
        }
        public void mouseExited(MouseEvent e)
        {
        }
    }

    private void createMenuBar()
    {
        JMenuItem fileMenuOpen = new JMenuItem("Open");
        fileMenuOpen.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                Project project = Project.from_File(selectDestinationFile("Select Part", "", ".dat"), config.defaultColorCode);
                Model.getInstance().merge(project);
            }
        });
        fileMenuOpen.setName("fileMenuOpen");

        JMenuItem fileMenuExit = new JMenuItem("Exit");
        fileMenuExit.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                System.exit(0);
            }
        });
        fileMenuExit.setName("fileMenuExit");

        JMenu fileMenu = new JMenu("File");
        fileMenu.setName("fileMenu");
        fileMenu.add(fileMenuOpen);
        fileMenu.add(fileMenuExit);


        JMenuItem aboutMenuItem = new JMenuItem("About");
        aboutMenuItem.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                showAboutBox();
            }
        });
        aboutMenuItem.setName("aboutMenuItem");


        JMenu helpMenu = new JMenu("Help");
        helpMenu.setName("helpMenu");
        helpMenu.add(aboutMenuItem);


        menuBar = new JMenuBar();
        menuBar.setName("menuBar");
        menuBar.add(fileMenu);
        menuBar.add(helpMenu);
    }

    public void showAboutBox()
    {
        log.info("HELP");
//        if (aboutBox == null)
//        {
//            JFrame mainFrame = Main.getApplication().getMainFrame();
//            aboutBox = new LOCADAboutBox(mainFrame);
//            aboutBox.setLocationRelativeTo(mainFrame);
//        }
//        Main.getApplication().show(aboutBox);
    }

    public static class FileFilter extends javax.swing.filechooser.FileFilter
    {
        private final String fileExtension;

        public FileFilter(String fileExtension)
        {
            this.fileExtension = fileExtension;
        }

        public boolean accept(File file)
        {
            if (file == null)
            {
                throw new IllegalArgumentException("nullValue.FileIsNull");
            }

            return file.isDirectory() || file.getName().toLowerCase().endsWith(fileExtension);
        }

        public String getDescription()
        {
            return "LDraw Model (ldr)";
        }
    }

    private File selectDestinationFile(String title, String filename, String fileExension)
    {
        File destFile = null;

        if (this.fileChooser == null)
        {
            this.fileChooser = new JFileChooser();
            this.fileChooser.setCurrentDirectory(new File(Configuration.ldrawDir));
            this.fileChooser.addChoosableFileFilter(new FileFilter(fileExension));
        }

        this.fileChooser.setDialogTitle(title);
        this.fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
        this.fileChooser.setMultiSelectionEnabled(false);
        this.fileChooser.setDialogType(JFileChooser.SAVE_DIALOG);


        this.fileChooser.setName(filename);

        int status = this.fileChooser.showSaveDialog(null);
        if (status == JFileChooser.APPROVE_OPTION)
        {
            destFile = this.fileChooser.getSelectedFile();
            if (!destFile.getName().endsWith(fileExension))
            {
                destFile = new File(destFile.getPath());
            }
        }
        return destFile;
    }
}
