
package com.locad.model;

import com.locad.model.annotations.*;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import org.apache.log4j.Logger;

public class Model
{
    private static final int INITIAL_CAPACITY = 100;
    private static Model instance = new Model();
    private ModelListenerDelegate delegate;
    private List<ContainerListener> containerListeners = null;
    private Logger log = null;
    private ConcurrentHashMap<Class, ConcurrentHashMap<String, ModelObject>> storage = new ConcurrentHashMap<Class, ConcurrentHashMap<String, ModelObject>>();
    private Hashtable<Class, Hashtable<Class, List<Method>>> annotations = new Hashtable<Class, Hashtable<Class, List<Method>>>();

    private Model()
    {
        log = Logger.getLogger(getClass().getName());
        delegate = new ModelListenerDelegate();
        containerListeners = new LinkedList<ContainerListener>();
    }

    public static Model getInstance()
    {
        return instance;
    }

    private <T extends ModelObject> void callMethodWithAnnotation(T t, Class annotationClass)
    {
        try
        {
            if (annotations.containsKey(t.getClass()))
            {
                if (annotations.get(t.getClass()).containsKey(annotationClass))
                {
                    for (Method method : annotations.get(t.getClass()).get(annotationClass))
                    {
                        method.invoke(t);
                    }
                }
                else
                {
                    annotations.get(t.getClass()).put(annotationClass, new LinkedList<Method>());
                    for (Method method : t.getClass().getMethods())
                    {
                        if (method.isAnnotationPresent(annotationClass))
                        {
                            annotations.get(t.getClass()).get(annotationClass).add(method);
                            method.invoke(t);
                        }
                    }
                }
            }
            else
            {
                annotations.put(t.getClass(), new Hashtable<Class, List<Method>>());
                annotations.get(t.getClass()).put(annotationClass, new LinkedList<Method>());
                for (Method method : t.getClass().getMethods())
                {
                    if (method.isAnnotationPresent(annotationClass))
                    {
                        annotations.get(t.getClass()).get(annotationClass).add(method);
                        method.invoke(t);
                    }
                }
            }
        }
        catch (IllegalAccessException ex)
        {
        }
        catch (InvocationTargetException ex)
        {
        }
    }

    public synchronized <T extends ModelObject> void merge(List<T> list)
    {
        List<T> updated = new LinkedList<T>();
        List<T> added = new LinkedList<T>();

        for (T t : list)
        {
            ConcurrentHashMap<String, ModelObject> type = storage.get(t.getClass());

            if (type == null)
            {
                log.info("Creating container for " + t.getClass().getName() + "\"");
                type = new ConcurrentHashMap<String, ModelObject>(INITIAL_CAPACITY);
                storage.put(t.getClass(), type);

                for (ContainerListener l : containerListeners)
                {
                    l.containerAdded(t.getClass());
                }
            }

            if (type.containsKey(t.getUuid()))
            {
                log.info("Updating " + t.getClass().getName() + "\"");
                callMethodWithAnnotation(type.get(t.getUuid()), PreUpdate.class);
                type.get(t.getUuid()).update(t);
                callMethodWithAnnotation(type.get(t.getUuid()), PostUpdate.class);
                updated.add((T) type.get(t.getUuid()));
            }
            else
            {
                callMethodWithAnnotation(t, PreAdd.class);
                type.put(t.getUuid(), t);
                callMethodWithAnnotation(t, PostAdd.class);
                added.add(t);
            }
        }

        delegate.notifyUpdate(updated);
        delegate.notifyAdd(added);
    }

    public synchronized <T extends ModelObject> void merge(T t)
    {
        ConcurrentHashMap<String, ModelObject> type = storage.get(t.getClass());

        if (type == null)
        {
            log.info("Creating container for " + t.getClass().getName() + "\"");
            type = new ConcurrentHashMap<String, ModelObject>(INITIAL_CAPACITY);
            storage.put(t.getClass(), type);

            for (ContainerListener l : containerListeners)
            {
                l.containerAdded(t.getClass());
            }
        }

        if (type.containsKey(t.getUuid()))
        {
            log.debug("Updating " + t.getClass().getName() + "\"");
            callMethodWithAnnotation(type.get(t.getUuid()), PreUpdate.class);
            type.get(t.getUuid()).update(t);
            callMethodWithAnnotation(type.get(t.getUuid()), PostUpdate.class);
            delegate.notifyUpdate(type.get(t.getUuid()));
        }
        else
        {
            callMethodWithAnnotation(t, PreAdd.class);
            type.put(t.getUuid(), t);
            callMethodWithAnnotation(t, PostAdd.class);
            delegate.notifyAdd(t);
        }
    }

    public synchronized <T extends ModelObject> void remove(Class<T> cl, T t)
    {
        log.info("Removing " + t.getClass().getName() + "\"");
        callMethodWithAnnotation(t, PreRemove.class);
        storage.get(cl).remove(t.getUuid());
        callMethodWithAnnotation(t, PostRemove.class);
        delegate.notifyRemove(t);
    }

    public synchronized <T extends ModelObject> void remove(List<T> list)
    {
        for (T t : list)
        {
            log.info("Removing " + t.getClass().getName() + "\"");
            callMethodWithAnnotation(t, PreRemove.class);
            storage.get(t.getClass()).remove(t.getUuid());
            callMethodWithAnnotation(t, PostRemove.class);
        }
        delegate.notifyRemove(list);
    }

    public <T extends ModelObject> List<T> findIncludingProgeny(Class<T> cl, String uuid) throws EntityNotFoundException, NonUniqueIDException
    {
        List<T> keys = new LinkedList<T>();

        for (Class clazz : storage.keySet())
        {
            if (cl.isAssignableFrom(clazz) && storage.keySet().contains(clazz) && storage.get(clazz).keySet().contains(uuid))
            {
                keys.add((T) storage.get(clazz).get(uuid));
            }
        }

        if (keys.isEmpty())
        {
            throw new EntityNotFoundException("Could not find class \"" + cl.getName() + "\" with UUID \"" + uuid + "\"");
        }

        return keys;
    }

    public <T extends ModelObject> T find(Class<T> cl, String uuid) throws EntityNotFoundException
    {
        if (!storage.keySet().contains(cl))
        {
            throw new EntityNotFoundException("Could not find class \"" + cl.getName() + "\"");
        }

        if (!storage.get(cl).keySet().contains(uuid))
        {
            throw new EntityNotFoundException("Could not find class \"" + cl.getName() + "\" with UUID \"" + uuid + "\"");
        }

        return (T) storage.get(cl).get(uuid);
    }

    public synchronized <T extends ModelObject> Collection<T> getAllInstances(Class<T> cl) throws EntityNotFoundException
    {
        if (!storage.keySet().contains(cl))
        {
            throw new EntityNotFoundException("Could not find class \"" + cl.getName() + "\"");
        }

        return (Collection<T>) storage.get(cl).values();
    }

    public synchronized <T extends ModelObject> Collection<T> getAllInstancesIncludingProgeny(Class<T> cl) throws EntityNotFoundException
    {
        Collection<T> objects = new LinkedList<T>();

        for (Class clazz : storage.keySet())
        {
            if (cl.isAssignableFrom(clazz) && storage.keySet().contains(clazz))
            {
                objects.addAll((Collection<T>) storage.get(clazz).values());
            }
        }

        if (objects.size() <= 0)
        {
            throw new EntityNotFoundException("Could not find class \"" + cl.getName() + "\"");
        }

        return objects;
    }

    public Set<String> getKeysIncludingProgeny(Class cl) throws EntityNotFoundException
    {
        Set<String> keys = new TreeSet<String>();
        for (Class clazz : storage.keySet())
        {
            if (cl.isAssignableFrom(clazz))
            {
                keys.addAll(storage.get(clazz).keySet());
            }
        }
        if (keys.isEmpty())
        {
            throw new EntityNotFoundException("Could not find class \"" + cl.getName() + "\"");
        }
        return keys;
    }

    public Set<String> getKeys(Class cl) throws EntityNotFoundException
    {
        if (!storage.keySet().contains(cl))
        {
            throw new EntityNotFoundException("Could not find class \"" + cl.getName() + "\"");
        }

        return storage.get(cl).keySet();
    }

    public <T extends ModelObject> List<T> findByField(Class<T> cl, String field, Object value) throws EntityNotFoundException, NoSuchFieldException
    {
        if (!storage.keySet().contains(cl))
        {
            throw new EntityNotFoundException("Could not find class \"" + cl.getName() + "\"");
        }

        List<T> list = new LinkedList<T>();

        Method m = null;
        try
        {
            m = cl.getMethod("get" + Character.toString(field.charAt(0)).toUpperCase() + field.substring(1));
        }
        catch (NoSuchMethodException ex)
        {
            throw new NoSuchFieldException();
        }

        for (ModelObject obj : storage.get(cl).values())
        {
            try
            {
                if (value.equals(m.invoke(obj)))
                {
                    list.add((T) obj);
                }
            }
            catch (IllegalAccessException ex)
            {
                throw new NoSuchFieldException();
            }
            catch (InvocationTargetException ex)
            {
                throw new NoSuchFieldException();
            }
        }

        if (list.isEmpty())
        {
            throw new EntityNotFoundException("Could not find class \"" + cl.getName() + "\" with field \"" + field + "\" valued \"" + value + "\"");
        }

        return list;
    }

    public <T extends ModelObject> List<T> findByFieldIncludingProgeny(Class<T> cl, String field, Object value) throws EntityNotFoundException, NoSuchFieldException
    {

        List<T> list = new LinkedList<T>();

        for (Class clazz : storage.keySet())
        {
            if (cl.isAssignableFrom(clazz))
            {

                Method m = null;
                try
                {
                    m = clazz.getMethod("get" + Character.toString(field.charAt(0)).toUpperCase() + field.substring(1));
                }
                catch (NoSuchMethodException ex)
                {
                    throw new NoSuchFieldException();
                }

                for (ModelObject obj : storage.get(clazz).values())
                {
                    try
                    {
                        if (value.equals(m.invoke(obj)))
                        {
                            list.add((T) obj);
                        }
                    }
                    catch (IllegalAccessException ex)
                    {
                        throw new NoSuchFieldException();
                    }
                    catch (InvocationTargetException ex)
                    {
                        throw new NoSuchFieldException();
                    }
                }
            }
        }


        if (list.isEmpty())
        {
            throw new EntityNotFoundException("Could not find class \"" + cl.getName() + "\" with field \"" + field + "\" valued \"" + value + "\"");
        }

        return list;
    }

    public <T extends ModelObject> List<T> findByContains(Class<T> cl, String field, Object value) throws EntityNotFoundException, NoSuchFieldException
    {
        if (!storage.keySet().contains(cl))
        {
            throw new EntityNotFoundException("Could not find class \"" + cl.getName() + "\"");
        }

        List<T> list = new LinkedList<T>();

        Method m = null;
        try
        {
            m = cl.getMethod("get" + Character.toString(field.charAt(0)).toUpperCase() + field.substring(1));
            m.getReturnType().asSubclass(Collection.class);
        }
        catch (NoSuchMethodException ex)
        {
            throw new NoSuchFieldException();
        }
        catch (ClassCastException ex)
        {
            throw new NoSuchFieldException();
        }

        for (ModelObject obj : storage.get(cl).values())
        {
            try
            {
                Collection c = (Collection) m.invoke(obj);

                if (c.contains(value))
                {
                    list.add((T) obj);
                }
            }
            catch (IllegalAccessException ex)
            {
                throw new NoSuchFieldException();
            }
            catch (InvocationTargetException ex)
            {
                throw new NoSuchFieldException();
            }
        }

        if (list.isEmpty())
        {
            throw new EntityNotFoundException("Could not find class \"" + cl.getName() + "\" with collection \"" + field + "\" that contains \"" + value + "\"");
        }

        return list;
    }

    public <T extends ModelObject> List<T> findByContainsIncludingProgeny(Class<T> cl, String field, Object value) throws EntityNotFoundException, NoSuchFieldException
    {
        List<T> list = new LinkedList<T>();

        for (Class clazz : storage.keySet())
        {
            if (cl.isAssignableFrom(clazz))
            {

                Method m = null;
                try
                {
                    m = clazz.getMethod("get" + Character.toString(field.charAt(0)).toUpperCase() + field.substring(1));
                    m.getReturnType().asSubclass(Collection.class);
                }
                catch (NoSuchMethodException ex)
                {
                    throw new NoSuchFieldException();
                }
                catch (ClassCastException ex)
                {
                    throw new NoSuchFieldException();
                }

                for (ModelObject obj : storage.get(clazz).values())
                {
                    try
                    {
                        Collection c = (Collection) m.invoke(obj);

                        if (c.contains(value))
                        {
                            list.add((T) obj);
                        }
                    }
                    catch (IllegalAccessException ex)
                    {
                        throw new NoSuchFieldException();
                    }
                    catch (InvocationTargetException ex)
                    {
                        throw new NoSuchFieldException();
                    }
                }
            }
        }

        if (list.isEmpty())
        {
            throw new EntityNotFoundException("Could not find class \"" + cl.getName() + "\" with collection \"" + field + "\" that contains \"" + value + "\"");
        }

        return list;
    }

    public void addListener(Class cl, ModelListener listener)
    {
        if (!ModelObject.class.isAssignableFrom(cl))
        {
            log.error(listener.getClass() + " attempting to register listener on a non model object!");
        }
        else
        {
            delegate.addListener(cl, listener);

            for (Class clazz : storage.keySet())
            {
                if (cl.isAssignableFrom(clazz))
                {
                    for (String key : storage.get(clazz).keySet())
                    {
                        listener.add(storage.get(clazz).get(key));
                    }
                }
            }
        }
    }

    public void removeListener(Class cl, ModelListener listener)
    {
        delegate.removeListener(cl, listener);
    }

    public void clear()
    {
        for (Class cl : storage.keySet())
        {
            for (String uuid : storage.get(cl).keySet())
            {
                delegate.notifyRemove(storage.get(cl).get(uuid));
            }

            storage.get(cl).clear();
        }
    }

    public void clear(Class cl)
    {
        for (String uuid : storage.get(cl).keySet())
        {
            delegate.notifyRemove(storage.get(cl).get(uuid));
        }

        ConcurrentHashMap m = storage.get(cl);
        if (m != null)
        {
            m.clear();
        }
    }

    public Class getClassOf(String uuid)
    {
        for (Class cl : storage.keySet())
        {
            if (storage.get(cl).containsKey(uuid))
            {
                return cl;
            }
        }
        return null;
    }

    public void addContainerListener(ContainerListener listener)
    {
        containerListeners.add(listener);
        for (Class cl : storage.keySet())
        {
            listener.containerAdded(cl);
        }
    }

    public void removeContainerListener(ContainerListener listener)
    {
        containerListeners.remove(listener);
    }
}
