
package com.locad.model;

import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadFactory;
import org.apache.log4j.Logger;

public class ModelListenerDelegate
{
    private Logger log;
    private ExecutorService pool = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors() * 4, new CustomThreadFactory());
    private final ConcurrentHashMap<Class, List<ModelListener>> listeners = new ConcurrentHashMap<Class, List<ModelListener>>();

    public ModelListenerDelegate()
    {
        log = Logger.getLogger(getClass().getName());
    }

    public void addListener(Class cl, ModelListener l)
    {
        synchronized (listeners)
        {
            if (listeners.get(cl) == null)
            {
                listeners.put(cl, new LinkedList<ModelListener>());
            }
        }

        synchronized (listeners.get(cl))
        {
            listeners.get(cl).add(l);
        }
    }

    public void removeListener(Class cl, ModelListener l)
    {
        if (listeners.get(cl) != null)
        {
            synchronized (listeners.get(cl))
            {
                listeners.get(cl).remove(l);
            }
        }
    }

    public <T extends ModelObject> void notifyAdd(final List<T> list)
    {
        notifyAdd(list, listeners);
    }

    private <T extends ModelObject> void notifyAdd(final List<T> list, ConcurrentHashMap<Class, List<ModelListener>> listeners)
    {
        for (Class c : listeners.keySet())
        {
            if (list.size() > 0 && c.isAssignableFrom(list.get(0).getClass()))
            {
                synchronized (listeners.get(c))
                {
                    for (ModelListener l : listeners.get(c))
                    {
                        final ModelListener listener = l;
                        pool.execute(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                listener.add(list);
                            }
                        });
                    }
                }
            }
        }
    }

    public <T extends ModelObject> void notifyAdd(final T t)
    {
        notifyAdd(t, listeners);
    }

    private <T extends ModelObject> void notifyAdd(final T t, ConcurrentHashMap<Class, List<ModelListener>> listeners)
    {
        for (Class c : listeners.keySet())
        {
            if (c.isAssignableFrom(t.getClass()))
            {
                synchronized (listeners.get(c))
                {
                    for (ModelListener l : listeners.get(c))
                    {
                        final ModelListener listener = l;
                        pool.execute(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                listener.add(t);
                            }
                        });
                    }
                }
            }
        }
    }

    public <T extends ModelObject> void notifyRemove(final List<T> list)
    {
        notifyRemove(list, listeners);
    }

    private <T extends ModelObject> void notifyRemove(final List<T> list, ConcurrentHashMap<Class, List<ModelListener>> listeners)
    {
        for (Class c : listeners.keySet())
        {
            if (list.size() > 0 && c.isAssignableFrom(list.get(0).getClass()))
            {
                synchronized (listeners.get(c))
                {
                    for (ModelListener l : listeners.get(c))
                    {
                        final ModelListener listener = l;
                        pool.execute(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                listener.remove(list);
                            }
                        });
                    }
                }
            }
        }
    }

    public <T extends ModelObject> void notifyRemove(final T t)
    {
        notifyRemove(t, listeners);
    }

    private <T extends ModelObject> void notifyRemove(final T t, ConcurrentHashMap<Class, List<ModelListener>> listeners)
    {
        for (Class c : listeners.keySet())
        {
            if (c.isAssignableFrom(t.getClass()))
            {
                synchronized (listeners.get(c))
                {
                    for (ModelListener l : listeners.get(c))
                    {
                        final ModelListener listener = l;
                        pool.execute(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                listener.remove(t);
                            }
                        });
                    }
                }
            }
        }
    }

    public <T extends ModelObject> void notifyUpdate(final List<T> list)
    {
        notifyUpdate(list, listeners);
    }

    private <T extends ModelObject> void notifyUpdate(final List<T> list, ConcurrentHashMap<Class, List<ModelListener>> listeners)
    {
        for (Class c : listeners.keySet())
        {
            if (list.size() > 0 && list.get(0) != null && c.isAssignableFrom(list.get(0).getClass()))
            {
                synchronized (listeners.get(c))
                {
                    for (ModelListener l : listeners.get(c))
                    {
                        final ModelListener listener = l;
                        pool.execute(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                listener.update(list);
                            }
                        });
                    }
                }
            }
        }
    }

    public <T extends ModelObject> void notifyUpdate(final T t)
    {
        notifyUpdate(t, listeners);
    }

    private <T extends ModelObject> void notifyUpdate(final T t, ConcurrentHashMap<Class, List<ModelListener>> listeners)
    {
        for (Class c : listeners.keySet())
        {
            if (c.isAssignableFrom(t.getClass()))
            {
                synchronized (listeners.get(c))
                {
                    for (ModelListener l : listeners.get(c))
                    {
                        final ModelListener listener = l;
                        pool.execute(new Runnable()
                        {
                            @Override
                            public void run()
                            {
                                listener.update(t);
                            }
                        });
                    }
                }
            }
        }
    }

    private class CustomThreadFactory implements ThreadFactory
    {
        int num = 1;
        public Thread newThread(Runnable r)
        {
            Thread thread = new Thread(r, "ModelNotify-Pool-" + num);
            num++;
            return thread;
        }
    }
}
