
package com.locad.model;

import java.io.Serializable;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.Collection;
import org.apache.log4j.Logger;

public class ModelObject implements Serializable
{
    protected final String uuid;
    protected transient Logger log;

    public ModelObject(String uuid)
    {
        log = Logger.getLogger(getClass());

        this.uuid = uuid;
    }

    public String getUuid()
    {
        return uuid;
    }

    @Override
    public String toString()
    {
        String ret = getClass().getName() + "\n";

        Method[] thisMethods = this.getClass().getMethods();
        for (int i = 0; i < thisMethods.length; ++i)
        {
            String thisName = thisMethods[i].getName();

            if (!thisName.startsWith("get"))
            {
                continue;
            }

            try
            {
                Object obj = thisMethods[i].invoke(this);
                if (obj == null)
                {
                    ret += thisName.substring(3) + " : null " + "\n";
                }
                else if (obj instanceof ModelObject)
                {
                    ret += thisName.substring(3) + " : " + ((ModelObject) obj).toString() + "\n";
                }
                else
                {
                    ret += thisName.substring(3) + " : " + obj.toString() + "\n";
                }
            }
            catch (IllegalAccessException ex)
            {
                log.error(ex);
            }
            catch (InvocationTargetException ex)
            {
                log.error(ex);
            }
        }

        return ret;
    }

    public synchronized void update(Object obj)
    {
        if (!obj.getClass().getName().equals(this.getClass().getName()))
        {
            return;
        }

        String thisName;
        String thatName;

        Method[] thisMethods = this.getClass().getMethods();
        Method[] thatMethods = obj.getClass().getMethods();

        // Handle updating get methods that return objects of type Updatable
        for (int i = 0; i < thisMethods.length; ++i)
        {
            thisName = thisMethods[i].getName();

            if (thisName.startsWith("get"))
            {
                if (Updatable.class.isAssignableFrom(thisMethods[i].getReturnType()))
                {
                    for (int j = 0; j < thatMethods.length; ++j)
                    {
                        thatName = thatMethods[j].getName();
                        if (equals(thisName.toCharArray(), 3, thatName.toCharArray(), 3))
                        {
                            try
                            {
                                Object ret = thatMethods[j].invoke(obj);
                                Object upd = thisMethods[i].invoke(this);

                                if (((Updatable) upd).needsUpdate(ret))
                                {
                                    ((Updatable) upd).update(ret);
                                }
                            }
                            catch (IllegalAccessException ex)
                            {
                                log.error(ex);
                            }
                            catch (InvocationTargetException ex)
                            {
                                log.error(ex);
                            }
                        }
                    }
                }
            }
        }

        // Handle the get/set pairs for non "Updatable" objects
        for (int i = 0; i < thisMethods.length; ++i)
        {
            thisName = thisMethods[i].getName();

            if (!thisName.startsWith("set"))
            {
                continue;
            }

            for (int j = 0; j < thatMethods.length; ++j)
            {
                thatName = thatMethods[j].getName();

                if (!thatName.startsWith("get") && !thatName.startsWith("is"))
                {
                    continue;
                }

                if (equals(thisName.toCharArray(), 3, thatName.toCharArray(), 3) || equals(thisName.toCharArray(), 3, thatName.toCharArray(), 2))
                {
                    // Found the proper get/set pair
                    try
                    {
                        Object ret = thatMethods[j].invoke(obj);
                        if (checkParameter(ret))
                        {
                            thisMethods[i].invoke(this, ret);
                        }
                    }
                    catch (IllegalAccessException ex)
                    {
                        log.error(ex);
                    }
                    catch (InvocationTargetException ex)
                    {
                        log.error(ex);
                    }

                    break;
                }
            }
        }
    }

    protected boolean checkParameter(Object ret)
    {
        if (ret == null)
        {
            return false;
        }
        if (ret.toString().equals(""))
        {
            return false;
        }
        if (ret.toString().equalsIgnoreCase("NO STATEMENT"))
        {
            return false;
        }
        if (ret.toString().equalsIgnoreCase("NO_STATEMENT"))
        {
            return false;
        }
        if (ret.toString().equalsIgnoreCase("NO STATEMENT UNKNOWN"))
        {
            return false;
        }
        if (ret.toString().equalsIgnoreCase("NO_STATEMENT_UNKNOWN"))
        {
            return false;
        }
        if (ret.toString().equalsIgnoreCase("NO_STATEMENT_2"))
        {
            return false;
        }
        if (ret instanceof Double)
        {
            if (Double.isNaN((Double) ret))
            {
                return false;
            }
        }
        if (ret instanceof Integer)
        {
            if (((Integer) ret).intValue() == Integer.MIN_VALUE)
            {
                return false;
            }
        }
        if (ret instanceof Float)
        {
            if (Float.isNaN((Float) ret))
            {
                return false;
            }
        }
        if (ret instanceof Collection)
        {
            if (((Collection) ret).isEmpty())
            {
                return false;
            }
        }

        return true;
    }

    protected boolean checkHistoricalParameter(Object ret)
    {
        if (ret == null)
        {
            return false;
        }
        if (ret.toString().equals(""))
        {
            return false;
        }
        if (ret instanceof Double)
        {
            if (Double.isNaN((Double) ret))
            {
                return false;
            }
        }
        if (ret instanceof Integer)
        {
            if (((Integer) ret).intValue() == Integer.MIN_VALUE)
            {
                return false;
            }
        }
        if (ret instanceof Float)
        {
            if (Float.isNaN((Float) ret))
            {
                return false;
            }
        }
        if (ret instanceof Collection)
        {
            if (((Collection) ret).isEmpty())
            {
                return false;
            }
        }

        return true;
    }

    private boolean equals(char[] str1, int offset1, char[] str2, int offset2)
    {
        if (str1.length - offset1 != str2.length - offset2)
        {
            return false;
        }

        for (int i = 0; i < (str1.length - offset1) && i < (str2.length - offset2); ++i)
        {
            if (str1[i + offset1] != str2[i + offset2])
            {
                return false;
            }
        }

        return true;
    }
}
