
package com.locad.viewport;

import com.locad.enumerations.Enumerations.MaterialTypes;
import com.locad.enumerations.Enumerations.ViewAngleModes;
import com.locad.geometry.*;
import com.locad.ldraw.PartEmbeded;
import com.locad.ldraw.PartsLibrary;
import com.locad.ldraw.Project;
import com.locad.main.Configuration;
import com.locad.main.Desktop;
import com.locad.model.Model;
import java.awt.Point;
import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.glu.GLU;
import org.apache.log4j.Logger;

public class Renderer implements GLEventListener
{
    private final Logger log;

    private final Configuration config = Configuration.getInstance();
    private final PartsLibrary partsLib = PartsLibrary.getInstance();

    private ViewAngleModes viewAngle = ViewAngleModes.PARTS_BAG;

    private GLU glu = new GLU();
    private GL gl;
    private Project project;
    private final Camera camera;

    private float aspect = 0;
    private float width = 0;
    private float height = 0;

    public int scene_dl = -1;
    
    public Renderer(Camera camera, ViewAngleModes viewAngle)
    {
        log = Logger.getLogger(getClass().getName());

        this.camera = camera;
        this.viewAngle = viewAngle;
    }

    public Renderer(Project project, Camera camera, ViewAngleModes viewAngle)
    {
        log = Logger.getLogger(getClass().getName());

        this.project = project;
        this.project.getBoundingBox();
        this.camera = camera;
        this.viewAngle = viewAngle;
    }

    public void setProject(Project project)
    {
        this.project = project;
        this.resetCamera();
        scene_dl = -1;
    }

    public void resetCamera()
    {
        if (project != null)
        {
            project.getBoundingBox();
            camera.calcViewBox(project.getCentroid(), project.getMinPoint(), project.getMaxPoint(), aspect);
        }
    }

    public void setViewAngle(ViewAngleModes viewAngle)
    {
        this.viewAngle = viewAngle;
        camera.viewAngle = viewAngle;
        if (project != null)
        {
            project.getBoundingBox();
            camera.calcViewBox(project.getCentroid(), project.getMinPoint(), project.getMaxPoint(), aspect);
        }
    }

    public void init(GLAutoDrawable glDrawable)
    {
        gl = glDrawable.getGL();

        gl.glShadeModel(GL.GL_SMOOTH);

        gl.glClearColor(config.bgColor.getRed() / 255f, config.bgColor.getGreen() / 255f, config.bgColor.getBlue() / 255, config.bgColor.getAlpha() / 255f);
        gl.glClearDepth(1.0f);

        gl.glEnable(GL.GL_DEPTH_TEST);
        gl.glDepthFunc(GL.GL_LEQUAL);

        gl.glPolygonMode(GL.GL_FRONT, GL.GL_FILL);

        if (config.enableBFC)
        {
            gl.glCullFace(GL.GL_BACK);
            gl.glEnable(GL.GL_CULL_FACE);
        }

        if (config.enableLighting)
        {
//            gl.glShadeModel(mode);
//            gl.glShaderSource(shader, count, string, null);

            Light light0 = new Light(new float[] {0.4f, 0.4f, 0.4f, 1.0f}, new float[] {0.7f, 0.7f, 0.7f, 1.0f}, new float[] {0.3f, 0.3f, 0.3f, 1.0f}, new Vec4f(camera.pos.x, camera.pos.y, camera.pos.z, -1f));
            light0.ApplyToGL(gl, GL.GL_LIGHT0);
        }
        else
        {
            gl.glDisable(GL.GL_LIGHTING);
            gl.glDisable(GL.GL_BLEND);
        }

        createScene(gl);
    }

    private void createScene(GL gl)
    {
        if (gl != null && gl.glIsList(scene_dl))
        {
            gl.glDeleteLists(scene_dl, 1);
        }

        if (project == null || gl == null)
        {
            return;
        }

        scene_dl = gl.glGenLists(1);
        gl.glNewList(scene_dl, GL.GL_COMPILE);

        if (config.enableLighting || config.enableBFC)
        {
            gl.glLineWidth(0.01f);
            for (MaterialTypes materialType : MaterialTypes.values())
            {
                Material.setMaterialBFC(materialType, gl);
                project.addToGLwithBFC(gl, materialType);
            }
        }
        else
        {
            gl.glLineWidth(config.lineWidth);
            for (MaterialTypes materialType : MaterialTypes.values())
            {
                Material.setMaterial(materialType, gl);
                project.addToGL(gl, materialType);
            }
        }

        gl.glEndList();
        try
        {
            Thread.sleep(50);
        }
        catch (InterruptedException ex)
        {
        }

        if (gl.glIsList(scene_dl))
        {
            //log.info("Created Scene " + scene_dl);
        }
        else
        {
            log.error("Error creating scene " + scene_dl + " code " + gl.glGetError());
        }
    }

    public void display(GLAutoDrawable glDrawable)
    {
        if (scene_dl == -1)
        {
            createScene(gl);
        }
        
        gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);

        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();
        gl.glOrtho(camera.left, camera.right, camera.bottom, camera.top, camera.near, camera.far);
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glLoadIdentity();

        if (project != null)
        {
            glu.gluLookAt(camera.pos.x, camera.pos.y, camera.pos.z, camera.lookAt.x, camera.lookAt.y, camera.lookAt.z, camera.yaw, camera.roll, camera.pitch);

//            Point3f pos = camera.pos.copy();
//            pos.y -= 500;
//            gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, pos.toArray(), 0);

            // Render project by calling display list
            gl.glCallList(scene_dl);

//            if (config.enableLighting || config.enableBFC)
//            {
//                gl.glLineWidth(0.01f);
//                for (MaterialTypes materialType : MaterialTypes.values())
//                {
//                    Material.setMaterialBFC(materialType, gl);
//                    project.addToGLwithBFC(gl, materialType);
//                }
//            }
//            else
//            {
//                gl.glLineWidth(config.lineWidth);
//                for (MaterialTypes materialType : MaterialTypes.values())
//                {
//                    Material.setMaterial(materialType, gl);
//                    project.addToGL(gl, materialType);
//                }
//            }

            calcSelectionBox(gl);
        }

        gl.glFlush();
    }

    public Point3f mousePointToRenderLine(Point mousePoint)
    {
        float x = 0;
        float y = 0;

        // translate and normalize screen point to project point
        if (viewAngle == ViewAngleModes.FRONT)
        {
            x = mousePoint.x / width * (camera.right - camera.left) + camera.left + camera.lookAt.x;
            y = mousePoint.y / height * (camera.top - camera.bottom) + camera.bottom + camera.lookAt.y;
            return new Point3f(x, y, 0);
        }
        else if(viewAngle == ViewAngleModes.BACK)
        {
            x = (width - mousePoint.x) / width * (camera.right - camera.left) + camera.left + camera.lookAt.x;
            y = mousePoint.y / height * (camera.top - camera.bottom) + camera.bottom + camera.lookAt.y;
            return new Point3f(x, y, 0);
        }
        else if (viewAngle == ViewAngleModes.RIGHT)
        {
            x = mousePoint.x / width * (camera.right - camera.left) + camera.left + camera.lookAt.z;
            y = mousePoint.y / height * (camera.top - camera.bottom) + camera.bottom + camera.lookAt.y;
            return new Point3f(0, y, x);
        }
        else if (viewAngle == ViewAngleModes.LEFT)
        {
            x = (width -mousePoint.x) / width * (camera.right - camera.left) + camera.left + camera.lookAt.z;
            y = mousePoint.y / height * (camera.top - camera.bottom) + camera.bottom + camera.lookAt.y;
            return new Point3f(0, y, x);
        }
        else if(viewAngle == ViewAngleModes.TOP)
        {
            x = mousePoint.x / width * (camera.right - camera.left) + camera.left + camera.lookAt.x;
            y = (height - mousePoint.y) / height * (camera.top - camera.bottom) + camera.bottom + camera.lookAt.z;
            return new Point3f(x, 0, y);
        }
        else if(viewAngle == ViewAngleModes.BOTTOM)
        {
            x = (width - mousePoint.x) / width * (camera.right - camera.left) + camera.left + camera.lookAt.x;
            y = (height - mousePoint.y) / height * (camera.top - camera.bottom) + camera.bottom + camera.lookAt.z;
            return new Point3f(x, 0, y);
        }
        else
        {
            return null;
        }
    }

    public void doSelection(Point mousePoint, boolean clearSelections)
    {
        Point3f renderLine = mousePointToRenderLine(mousePoint);
        Shape minShape = project.getShape(renderLine, viewAngle);

        if (clearSelections)
        {
            for (Shape shape : project.getShapes())
            {
                shape.setIsSelected(false);
            }
        }

        if (minShape != null)
        {
            minShape.setIsSelected(true);
            Model.getInstance().merge(project);
        }

        Desktop.getInstance().repaint();
    }
    
    private void calcSelectionBox(GL gl)
    {
        Point3f minPoint = new Point3f(Float.MAX_VALUE, Float.MAX_VALUE, Float.MAX_VALUE);
        Point3f maxPoint = new Point3f(-Float.MAX_VALUE, -Float.MAX_VALUE, -Float.MAX_VALUE);

        for (Shape shape : project.getShapes())
        {
            Point3f min = new Point3f(Float.MAX_VALUE, Float.MAX_VALUE, Float.MAX_VALUE);
            Point3f max = new Point3f(-Float.MAX_VALUE, -Float.MAX_VALUE, -Float.MAX_VALUE);

            if (shape.getIsSelected())
            {
                if (shape instanceof PartEmbeded)
                {
                    PartEmbeded p = (PartEmbeded) shape;
                    partsLib.getPart(p.fileName).getBoundingBox(p.orientation.times(Matrix4f.identityMatrix()), min, max);
                    addSelectionBoxToGL(gl, min, max);
                }
                else
                {
                    shape.getBoundingBox(Matrix4f.identityMatrix(), min, max);
                    addSelectionBoxToGL(gl, min, max);
                }
                minPoint = minPoint.min(min);
                maxPoint = maxPoint.max(max);
            }
        }
        
        addSelectionBoxToGL(gl, minPoint, maxPoint);
    }

    private void addSelectionBoxToGL(GL gl, Point3f minPoint, Point3f maxPoint)
    {
        Point3f centroid = maxPoint.minus(maxPoint.minus(minPoint).div(2f));

        Line box1 = new Line(new Point3f(minPoint.x, minPoint.y, minPoint.z), new Point3f(maxPoint.x, minPoint.y, minPoint.z), (short) 7);
        Line box2 = new Line(new Point3f(maxPoint.x, minPoint.y, minPoint.z), new Point3f(maxPoint.x, maxPoint.y, minPoint.z), (short) 7);
        Line box3 = new Line(new Point3f(maxPoint.x, maxPoint.y, minPoint.z), new Point3f(minPoint.x, maxPoint.y, minPoint.z), (short) 7);
        Line box4 = new Line(new Point3f(minPoint.x, maxPoint.y, minPoint.z), new Point3f(minPoint.x, minPoint.y, minPoint.z), (short) 7);

        Line axis1 = new Line(new Point3f(centroid.x -1, centroid.y, centroid.z), new Point3f(centroid.x + 1, centroid.y, centroid.z), (short) 7);
        Line axis2 = new Line(new Point3f(centroid.x, centroid.y - 1, centroid.z), new Point3f(centroid.x, centroid.y + 1, centroid.z), (short) 7);

        gl.glLineWidth(Configuration.getInstance().lineWidth * 2f);

        if (viewAngle == ViewAngleModes.FRONT)
        {
            box1 = new Line(new Point3f(minPoint.x, minPoint.y, camera.pos.z+1), new Point3f(maxPoint.x, minPoint.y, camera.pos.z+1), (short) 7);
            box2 = new Line(new Point3f(maxPoint.x, minPoint.y, camera.pos.z+1), new Point3f(maxPoint.x, maxPoint.y, camera.pos.z+1), (short) 7);
            box3 = new Line(new Point3f(maxPoint.x, maxPoint.y, camera.pos.z+1), new Point3f(minPoint.x, maxPoint.y, camera.pos.z+1), (short) 7);
            box4 = new Line(new Point3f(minPoint.x, maxPoint.y, camera.pos.z+1), new Point3f(minPoint.x, minPoint.y, camera.pos.z+1), (short) 7);

            axis1 = new Line(new Point3f(centroid.x-1, centroid.y, camera.pos.z+1), new Point3f(centroid.x+1, centroid.y, camera.pos.z+1), (short) 7);
            axis2 = new Line(new Point3f(centroid.x, centroid.y-1, camera.pos.z+1), new Point3f(centroid.x, centroid.y+1, camera.pos.z+1), (short) 7);
        }
        else if (viewAngle == ViewAngleModes.BACK)
        {
            box1 = new Line(new Point3f(maxPoint.x, maxPoint.y, -camera.pos.z-1), new Point3f(minPoint.x, maxPoint.y, -camera.pos.z-1), (short) 7);
            box2 = new Line(new Point3f(minPoint.x, maxPoint.y, -camera.pos.z-1), new Point3f(minPoint.x, minPoint.y, -camera.pos.z-1), (short) 7);
            box3 = new Line(new Point3f(minPoint.x, minPoint.y, -camera.pos.z-1), new Point3f(maxPoint.x, minPoint.y, -camera.pos.z-1), (short) 7);
            box4 = new Line(new Point3f(maxPoint.x, minPoint.y, -camera.pos.z-1), new Point3f(maxPoint.x, maxPoint.y, -camera.pos.z-1), (short) 7);

            axis1 = new Line(new Point3f(centroid.x -1, centroid.y, -camera.pos.z-1), new Point3f(centroid.x + 1, centroid.y, -camera.pos.z-1), (short) 7);
            axis2 = new Line(new Point3f(centroid.x, centroid.y-1, -camera.pos.z-1), new Point3f(centroid.x, centroid.y+1, -camera.pos.z-1), (short) 7);
        }
        else if (viewAngle == ViewAngleModes.RIGHT)
        {
            box1 = new Line(new Point3f(-camera.pos.x-1, minPoint.y, minPoint.z), new Point3f(-camera.pos.x-1, minPoint.y, maxPoint.z), (short) 7);
            box2 = new Line(new Point3f(-camera.pos.x-1, minPoint.y, maxPoint.z), new Point3f(-camera.pos.x-1, maxPoint.y, maxPoint.z), (short) 7);
            box3 = new Line(new Point3f(-camera.pos.x-1, maxPoint.y, maxPoint.z), new Point3f(-camera.pos.x-1, maxPoint.y, minPoint.z), (short) 7);
            box4 = new Line(new Point3f(-camera.pos.x-1, maxPoint.y, minPoint.z), new Point3f(-camera.pos.x-1, minPoint.y, minPoint.z), (short) 7);

            axis1 = new Line(new Point3f(-camera.pos.x-1, centroid.y, centroid.z-1), new Point3f(-camera.pos.x-1, centroid.y, centroid.z+1), (short) 7);
            axis2 = new Line(new Point3f(-camera.pos.x-1, centroid.y-1, centroid.z), new Point3f(-camera.pos.x-1, centroid.y+1, centroid.z), (short) 7);
        }
        else if (viewAngle == ViewAngleModes.LEFT)
        {
            box1 = new Line(new Point3f(camera.pos.x+1, minPoint.y, minPoint.z), new Point3f(camera.pos.x+1, minPoint.y, maxPoint.z), (short) 7);
            box2 = new Line(new Point3f(camera.pos.x+1, minPoint.y, maxPoint.z), new Point3f(camera.pos.x+1, maxPoint.y, maxPoint.z), (short) 7);
            box3 = new Line(new Point3f(camera.pos.x+1, maxPoint.y, maxPoint.z), new Point3f(camera.pos.x+1, maxPoint.y, minPoint.z), (short) 7);
            box4 = new Line(new Point3f(camera.pos.x+1, maxPoint.y, minPoint.z), new Point3f(camera.pos.x+1, minPoint.y, minPoint.z), (short) 7);

            axis1 = new Line(new Point3f(camera.pos.x+1, centroid.y, centroid.z-1), new Point3f(camera.pos.x+1, centroid.y, centroid.z+1), (short) 7);
            axis2 = new Line(new Point3f(camera.pos.x+1, centroid.y-1, centroid.z), new Point3f(camera.pos.x+1, centroid.y+1, centroid.z), (short) 7);
        }
        else if (viewAngle == ViewAngleModes.TOP)
        {
            box1 = new Line(new Point3f(minPoint.x, camera.pos.y+2, minPoint.z), new Point3f(minPoint.x, camera.pos.y+2, maxPoint.z), (short) 7);
            box2 = new Line(new Point3f(minPoint.x, camera.pos.y+2, maxPoint.z), new Point3f(maxPoint.x, camera.pos.y+2, maxPoint.z), (short) 7);
            box3 = new Line(new Point3f(maxPoint.x, camera.pos.y+2, maxPoint.z), new Point3f(maxPoint.x, camera.pos.y+2, minPoint.z), (short) 7);
            box4 = new Line(new Point3f(maxPoint.x, camera.pos.y+2, minPoint.z), new Point3f(minPoint.x, camera.pos.y+2, minPoint.z), (short) 7);

            axis1 = new Line(new Point3f(centroid.x-1, camera.pos.y+2, centroid.z), new Point3f(centroid.x+1, camera.pos.y+2, centroid.z), (short) 7);
            axis2 = new Line(new Point3f(centroid.x, camera.pos.y+2, centroid.z-1), new Point3f(centroid.x, camera.pos.y+2, centroid.z+1), (short) 7);
        }
        else if (viewAngle == ViewAngleModes.BOTTOM)
        {
            box1 = new Line(new Point3f(minPoint.x, -camera.pos.y-1, minPoint.z), new Point3f(minPoint.x, -camera.pos.y-1, maxPoint.z), (short) 7);
            box2 = new Line(new Point3f(minPoint.x, -camera.pos.y-1, maxPoint.z), new Point3f(maxPoint.x, -camera.pos.y-1, maxPoint.z), (short) 7);
            box3 = new Line(new Point3f(maxPoint.x, -camera.pos.y-1, maxPoint.z), new Point3f(maxPoint.x, -camera.pos.y-1, minPoint.z), (short) 7);
            box4 = new Line(new Point3f(maxPoint.x, -camera.pos.y-1, minPoint.z), new Point3f(minPoint.x, -camera.pos.y-1, minPoint.z), (short) 7);

            axis1 = new Line(new Point3f(centroid.x-1, -camera.pos.y-1, centroid.z), new Point3f(centroid.x+1, -camera.pos.y-1, centroid.z), (short) 7);
            axis2 = new Line(new Point3f(centroid.x, -camera.pos.y-1, centroid.z-1), new Point3f(centroid.x, -camera.pos.y-1, centroid.z+1), (short) 7);
        }
        else
        {
            Line box5 = new Line(new Point3f(maxPoint.x, maxPoint.y, maxPoint.z), new Point3f(minPoint.x, maxPoint.y, maxPoint.z), (short) 7);
            Line box6 = new Line(new Point3f(minPoint.x, maxPoint.y, maxPoint.z), new Point3f(minPoint.x, minPoint.y, maxPoint.z), (short) 7);
            Line box7 = new Line(new Point3f(minPoint.x, minPoint.y, maxPoint.z), new Point3f(maxPoint.x, minPoint.y, maxPoint.z), (short) 7);
            Line box8 = new Line(new Point3f(maxPoint.x, minPoint.y, maxPoint.z), new Point3f(maxPoint.x, maxPoint.y, maxPoint.z), (short) 7);

            Line box9 = new Line(new Point3f(minPoint.x, minPoint.y, minPoint.z), new Point3f(minPoint.x, minPoint.y, maxPoint.z), (short) 7);
            Line box10 = new Line(new Point3f(maxPoint.x, minPoint.y, minPoint.z), new Point3f(maxPoint.x, minPoint.y, maxPoint.z), (short) 7);
            Line box11 = new Line(new Point3f(minPoint.x, maxPoint.y, minPoint.z), new Point3f(minPoint.x, maxPoint.y, maxPoint.z), (short) 7);
            Line box12 = new Line(new Point3f(maxPoint.x, maxPoint.y, maxPoint.z), new Point3f(maxPoint.x, maxPoint.y, minPoint.z), (short) 7);

            axis1 = new Line(new Point3f(centroid.x -1, centroid.y, centroid.z), new Point3f(centroid.x + 1, centroid.y, centroid.z), (short) 7);
            axis2 = new Line(new Point3f(centroid.x, centroid.y - 1, centroid.z), new Point3f(centroid.x, centroid.y + 1, centroid.z), (short) 7);            
            Line axis3 = new Line(new Point3f(centroid.x, centroid.y, centroid.z - 1), new Point3f(centroid.x, centroid.y, centroid.z + 1), (short) 7);

            box5.addToGL(gl, (short) 7, MaterialTypes.PLASTIC);
            box6.addToGL(gl, (short) 7, MaterialTypes.PLASTIC);
            box7.addToGL(gl, (short) 7, MaterialTypes.PLASTIC);
            box8.addToGL(gl, (short) 7, MaterialTypes.PLASTIC);

            box9.addToGL(gl, (short) 7, MaterialTypes.PLASTIC);
            box10.addToGL(gl, (short) 7, MaterialTypes.PLASTIC);
            box11.addToGL(gl, (short) 7, MaterialTypes.PLASTIC);
            box12.addToGL(gl, (short) 7, MaterialTypes.PLASTIC);

            axis3.addToGL(gl, (short) 7, MaterialTypes.PLASTIC);
        }

        box1.addToGL(gl, (short) 7, MaterialTypes.PLASTIC);
        box2.addToGL(gl, (short) 7, MaterialTypes.PLASTIC);
        box3.addToGL(gl, (short) 7, MaterialTypes.PLASTIC);
        box4.addToGL(gl, (short) 7, MaterialTypes.PLASTIC);

        axis1.addToGL(gl, (short) 7, MaterialTypes.PLASTIC);
        axis2.addToGL(gl, (short) 7, MaterialTypes.PLASTIC);
        

        gl.glLineWidth(Configuration.getInstance().lineWidth);
    }

    public void displayChanged(GLAutoDrawable glDrawable, boolean modeChanged, boolean deviceChanged)
    {
        //init(glDrawable);
    }

    public void reshape(GLAutoDrawable glDrawable, int x, int y, int w, int h)
    {
        aspect = (float) w / (float) h;
        width = w;
        height = h;

        if (project != null)
        {
            camera.calcViewBox(project.getCentroid(), project.getMinPoint(), project.getMaxPoint(), aspect);
        }

        gl.glViewport(0, 0, w, h);
        gl.glMatrixMode(GL.GL_PROJECTION);
        gl.glLoadIdentity();
        gl.glOrtho(camera.left, camera.right, camera.bottom, camera.top, camera.near, camera.far);
        camera.setBounds((float) w, (float) h);
    }
}
